"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import os
import shutil
import sys
from functools import partial

import click
from tabulate import tabulate

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.build.helper import resolve_bundles_cache_directory
from syndicate.core.export.export_processor import export_specification
from syndicate.core.transform.transform_processor import generate_build_meta
from syndicate.core import initialize_connection, \
    initialize_project_state, initialize_signal_handling
from syndicate.core.build.artifact_processor import RUNTIME_NODEJS, \
    assemble_artifacts, RUNTIME_JAVA, RUNTIME_PYTHON, RUNTIME_SWAGGER_UI, \
    RUNTIME_DOTNET, RUNTIME_APPSYNC
from syndicate.core.build.bundle_processor import create_bundles_bucket, \
    load_bundle, upload_bundle_to_s3, if_bundle_exist, \
    remove_bundle_dir_locally
from syndicate.core.build.deployment_processor import \
    create_deployment_resources, remove_deployment_resources, \
    update_deployment_resources
from syndicate.core.build.meta_processor import create_meta
from syndicate.core.build.profiler_processor import (get_metric_statistics,
                                                     process_metrics)
from syndicate.core.build.warmup_processor import process_deploy_resources, \
    process_api_gw_resources, warm_upper, process_existing_api_gw_id, \
    process_inputted_api_gw_id
from syndicate.core.conf.validator import (JAVA_LANGUAGE_NAME,
                                           PYTHON_LANGUAGE_NAME,
                                           NODEJS_LANGUAGE_NAME,
                                           SWAGGER_UI_NAME,
                                           DOTNET_LANGUAGE_NAME, APPSYNC_NAME)
from syndicate.core.decorators import (check_deploy_name_for_duplicates,
                                       check_deploy_bucket_exists,
                                       check_bundle_deploy_names_for_existence,
                                       return_code_manager)
from syndicate.core.groups.generate import (generate,
                                            GENERATE_PROJECT_COMMAND_NAME,
                                            GENERATE_CONFIG_COMMAND_NAME)
from syndicate.core.groups.tags import tags
from syndicate.core.helper import (create_bundle_callback,
                                   handle_futures_progress_bar,
                                   resolve_path_callback, timeit,
                                   verify_bundle_callback, sync_lock,
                                   resolve_default_value, ValidRegionParamType,
                                   generate_default_bundle_name,
                                   resolve_and_verify_bundle_callback,
                                   param_to_lower, verbose_option,
                                   validate_incompatible_options,
                                   failed_status_code_on_exception,
                                   AliasedCommandsGroup, MultiWordOption,
                                   are_resource_types_valid)
from syndicate.core.project_state.project_state import (MODIFICATION_LOCK,
                                                        WARMUP_LOCK)
from syndicate.core.project_state.status_processor import project_state_status
from syndicate.core.project_state.sync_processor import sync_project_state
from syndicate.core.constants import TEST_ACTION, BUILD_ACTION, \
    DEPLOY_ACTION, UPDATE_ACTION, CLEAN_ACTION, SYNC_ACTION, ABORTED_STATUS, \
    STATUS_ACTION, WARMUP_ACTION, PROFILER_ACTION, ASSEMBLE_JAVA_MVN_ACTION, \
    ASSEMBLE_PYTHON_ACTION, ASSEMBLE_NODE_ACTION, ASSEMBLE_ACTION, \
    PACKAGE_META_ACTION, CREATE_DEPLOY_TARGET_BUCKET_ACTION, UPLOAD_ACTION, \
    COPY_BUNDLE_ACTION, EXPORT_ACTION, ASSEMBLE_SWAGGER_UI_ACTION, \
    ASSEMBLE_DOTNET_ACTION, ASSEMBLE_APPSYNC_ACTION, OK_RETURN_CODE, \
    FAILED_RETURN_CODE, ABORTED_RETURN_CODE, UPDATE_RESOURCE_TYPE_PRIORITY, \
    UNDERSCORE_CREATE_DEPLOY_TARGET_BUCKET_ACTION, \
    DEPLOY_RESOURCE_TYPE_PRIORITY, CLEAN_RESOURCE_TYPE_PRIORITY
from syndicate.exceptions import ProjectStateError
from syndicate import __version__

INIT_COMMAND_NAME = 'init'
SYNDICATE_PACKAGE_NAME = 'aws-syndicate'
HELP_PARAMETER_KEY = '--help'
commands_without_config = (
    INIT_COMMAND_NAME,
    GENERATE_PROJECT_COMMAND_NAME,
    GENERATE_CONFIG_COMMAND_NAME,
    HELP_PARAMETER_KEY
)

commands_without_state_sync = (
    CREATE_DEPLOY_TARGET_BUCKET_ACTION,
    UNDERSCORE_CREATE_DEPLOY_TARGET_BUCKET_ACTION
)

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


def _not_require_config(all_params):
    return any(item in commands_without_config for item in all_params)


def _not_require_state_sync(all_params):
    return any(item in commands_without_state_sync for item in all_params)


@click.group(name='syndicate', cls=AliasedCommandsGroup)
@return_code_manager
@click.version_option(version=__version__)
def syndicate():
    from syndicate.core import CONF_PATH
    if _not_require_config(sys.argv):
        pass
    elif CONF_PATH:
        USER_LOG.info('Configuration used: ' + CONF_PATH)
        initialize_connection()
        initialize_project_state(
            do_not_sync_state=_not_require_state_sync(sys.argv)
        )
        initialize_signal_handling()
    else:
        USER_LOG.error('Environment variable SDCT_CONF is not set! '
                       'Please verify that you have provided path to '
                       'correct config files '
                       'or execute `syndicate generate config` command.')
        sys.exit(FAILED_RETURN_CODE)


@syndicate.command(name=TEST_ACTION)
@return_code_manager
@click.option('--suite', default='unittest',
              type=click.Choice(['unittest', 'pytest', 'nose'],
                                case_sensitive=False),
              help='Supported testing frameworks. Possible options: unittest, '
                   'pytest, nose. Default value: unittest')
@click.option('--test-folder-name',
              cls=MultiWordOption, nargs=1, default='tests',
              help='Directory in the project that contains tests to run. '
                   'Default folder: tests')
@click.option('--errors-allowed',
              cls=MultiWordOption, is_flag=True,
              help='Flag to return successful response even if some tests '
                   'fail')
@click.option('--skip-tests',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to not run tests')
@verbose_option
@timeit(action_name=TEST_ACTION)
@failed_status_code_on_exception
def test(suite, test_folder_name, errors_allowed, skip_tests):
    """Discovers and runs tests inside python project configuration path."""
    if skip_tests:
        USER_LOG.info('Skipping tests...')
        return OK_RETURN_CODE

    USER_LOG.info('Running tests...')
    import subprocess
    from syndicate.core import CONFIG
    project_path = CONFIG.project_path
    test_folder = os.path.join(project_path, test_folder_name)
    if not os.path.exists(test_folder):
        msg = (f'Tests not found, \'{test_folder_name}\' folder is missing in '
               f'\'{project_path}\'.')
        USER_LOG.info(msg)
        return OK_RETURN_CODE
    test_lib_command_mapping = {
        'unittest': f'{sys.executable} -m unittest discover {test_folder} -v',
        'pytest': 'pytest --no-header -v',
        'nose': 'nosetests --verbose'
    }

    command = test_lib_command_mapping.get(suite)
    result = subprocess.run(command, cwd=project_path, shell=True,
                            capture_output=True, text=True)

    if result.returncode != OK_RETURN_CODE:
        _LOG.error(f'{result.stdout}\n{result.stderr}\n{"-" * 70}')
        if not errors_allowed:
            USER_LOG.error(
                'Some tests failed. See details in the log file. Exiting...')
            return result.returncode
        else:
            USER_LOG.warning('Some tests failed. See details in the log file.')
            return OK_RETURN_CODE
    else:
        _LOG.info(f'{result.stdout}\n{result.stderr}\n{"-" * 70}')
        USER_LOG.info('Tests passed.')
        return OK_RETURN_CODE


@syndicate.command(name=BUILD_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle to build. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%SZ')
@click.option('--force-upload', '-F',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to override existing bundle with the same name')
@click.option('--errors-allowed',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to continue building the bundle if any errors occur '
                   'while building dependencies or tests fail')
@click.option('--skip-tests',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to skip lambda tests')
@click.option('--refresh-cache',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to refresh the cache of dependencies. Currently, it '
                   'is only applicable to Python runtime dependencies.')
@verbose_option
@click.pass_context
@timeit(action_name=BUILD_ACTION)
@failed_status_code_on_exception
@check_deploy_bucket_exists
def build(ctx, bundle_name, force_upload, errors_allowed, skip_tests,
          refresh_cache):
    """
    Builds bundle of an application
    """
    if not force_upload:
        if if_bundle_exist(bundle_name=bundle_name):
            raise ProjectStateError(
                f'Bundle name \'{bundle_name}\' already exists in deploy '
                f'bucket. Please use another bundle name or delete the bundle'
            )
    remove_bundle_dir_locally(bundle_name, force_upload)

    test_code = ctx.invoke(test,
                           errors_allowed=errors_allowed,
                           skip_tests=skip_tests)
    if test_code != OK_RETURN_CODE:
        return test_code

    assemble_code = ctx.invoke(assemble,
                               bundle_name=bundle_name,
                               errors_allowed=errors_allowed,
                               skip_tests=skip_tests,
                               refresh_cache=refresh_cache,
                               is_chained=True)
    if assemble_code != OK_RETURN_CODE:
        return assemble_code

    meta_code = ctx.invoke(package_meta, bundle_name=bundle_name)
    if meta_code != OK_RETURN_CODE:
        return meta_code

    upload_code = ctx.invoke(upload,
                             bundle_name=bundle_name,
                             force_upload=force_upload)
    return upload_code


@syndicate.command(name='transform')
@return_code_manager
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, callback=resolve_default_value,
              help='Name of the bundle to transform. '
                   'Default value: name of the latest built bundle')
@click.option('--dsl', type=click.Choice(['CloudFormation', 'Terraform'],
                                         case_sensitive=False),
              callback=param_to_lower, multiple=True, required=True,
              help='Type of the IaC provider')
@click.option('--output-dir',
              cls=MultiWordOption,
              help='The directory where a transformed template will be saved')
@verbose_option
@timeit()
def transform(bundle_name, dsl, output_dir):
    """
    Transforms the meta-description of a bundle to a template
    compatible with the specified IaC provider
    """
    generate_build_meta(bundle_name=bundle_name,
                        dsl_list=dsl,
                        output_directory=output_dir)
    return OK_RETURN_CODE


@syndicate.command(name=DEPLOY_ACTION)
@return_code_manager
@sync_lock(lock_type=MODIFICATION_LOCK)
@timeit(action_name=DEPLOY_ACTION)
@failed_status_code_on_exception
@click.option('--deploy-name', '-d',
              cls=MultiWordOption, callback=resolve_default_value,
              help='Name of the deploy. Default value: name of the project')
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, callback=resolve_default_value,
              help='Name of the bundle to deploy. '
                   'Default value: name of the latest built bundle')
@click.option('--deploy-only-types', '-types',
              cls=MultiWordOption, multiple=True,
              type=str,
              help='Types of the resources to deploy')
@click.option('--deploy-only-resources', '-resources',
              cls=MultiWordOption, multiple=True,
              help='Names of the resources to deploy')
@click.option('--deploy-only-resources-path', '-path',
              cls=MultiWordOption, nargs=1, type=str,
              help='Path to file containing names of the resources to deploy')
@click.option('--excluded-resources', '-exresources',
              cls=MultiWordOption, multiple=True,
              help='Names of the resources to skip while deploy.')
@click.option('--excluded-resources-path', '-expath',
              cls=MultiWordOption, nargs=1, type=str,
              help='Path to file containing names of the resources to skip '
                   'while deploy')
@click.option('--excluded-types', '-extypes',
              cls=MultiWordOption, multiple=True, type=str,
              help='Types of the resources to skip while deploy')
@click.option('--continue-deploy',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to continue failed deploy')
@click.option('--replace-output',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to replace the existing deploy output')
@click.option('--rollback-on-error',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to automatically clean deployed resources if the'
                   ' deployment is unsuccessful')
@verbose_option
@check_deploy_name_for_duplicates
@check_deploy_bucket_exists
@check_bundle_deploy_names_for_existence()
def deploy(
        deploy_name: str,
        bundle_name: str,
        deploy_only_types: tuple | None = None,
        deploy_only_resources: tuple | None = None,
        deploy_only_resources_path: str | None = None,
        excluded_resources: tuple | None = None,
        excluded_resources_path: str | None = None,
        excluded_types: tuple | None = None,
        continue_deploy: bool = False,
        replace_output: bool = False,
        rollback_on_error: bool = False,
):
    """
    Deploys the application infrastructure
    """
    if not are_resource_types_valid(
            'excluded-types', excluded_types, DEPLOY_RESOURCE_TYPE_PRIORITY) or \
            not are_resource_types_valid(
                'deploy-only-types', deploy_only_types, DEPLOY_RESOURCE_TYPE_PRIORITY):
        return ABORTED_STATUS

    from syndicate.core import PROJECT_STATE
    PROJECT_STATE.current_bundle = bundle_name

    if deploy_only_resources_path and os.path.exists(
            deploy_only_resources_path):
        deploy_resources_list = json.load(open(deploy_only_resources_path))
        deploy_only_resources = tuple(
            set(deploy_only_resources + tuple(deploy_resources_list)))
        if deploy_only_resources:
            USER_LOG.info(
                f'Resources to deploy: {list(deploy_only_resources)}')

    if excluded_resources_path and os.path.exists(excluded_resources_path):
        excluded_resources_list = json.load(open(excluded_resources_path))
        excluded_resources = tuple(
            set(excluded_resources + tuple(excluded_resources_list)))
        if excluded_resources:
            USER_LOG.info(f'Resources to deploy: {list(excluded_resources)}')

    deploy_success = create_deployment_resources(
        deploy_name=deploy_name,
        bundle_name=bundle_name,
        deploy_only_types=deploy_only_types,
        deploy_only_resources=deploy_only_resources,
        excluded_resources=excluded_resources,
        excluded_types=excluded_types,
        continue_deploy=continue_deploy,
        replace_output=replace_output,
        rollback_on_error=rollback_on_error,
    )

    if deploy_success == ABORTED_STATUS:
        USER_LOG.warning('Deployment of resources has been aborted')
        return ABORTED_RETURN_CODE

    message = 'Backend resources were deployed{suffix}.'.format(
        suffix='' if deploy_success else (
            ' with errors. Rollback is enabled, resources from this '
            'deployment were cleaned'
            if rollback_on_error else ' with errors. See deploy output file'
        )
    )

    if not deploy_success:
        USER_LOG.warning(message)
        return FAILED_RETURN_CODE

    USER_LOG.info(message)
    return OK_RETURN_CODE


@syndicate.command(name=UPDATE_ACTION)
@return_code_manager
@sync_lock(lock_type=MODIFICATION_LOCK)
@timeit(action_name=UPDATE_ACTION)
@failed_status_code_on_exception
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, callback=resolve_default_value,
              help='Name of the bundle to update from. '
                   'Default value: name of the latest built bundle')
@click.option('--deploy-name', '-d',
              cls=MultiWordOption, callback=resolve_default_value,
              help='Name of the deploy. Default value: name of the project')
@click.option('--update-only-types', '-types',
              cls=MultiWordOption, multiple=True, type=str,
              help='Types of the resources to update')
@click.option('--update-only-resources',
              '-resources', cls=MultiWordOption, multiple=True,
              help='Names of the resources to update')
@click.option('--update-only-resources-path',
              '-path', cls=MultiWordOption, nargs=1,
              help='Path to file containing names of the resources to skip '
                   'while deploy')
@click.option('--excluded-resources', '-exresources',
              cls=MultiWordOption, multiple=True,
              help='Names of the resources to skip while update.')
@click.option('--excluded-resources-path',
              '-expath', cls=MultiWordOption, nargs=1,
              help='Path to file containing names of the resources to skip '
                   'while update')
@click.option('--excluded-types', '-extypes',
              cls=MultiWordOption, multiple=True, type=str,
              help='Types of the resources to skip while update')
@click.option('--replace-output', nargs=1,
              cls=MultiWordOption, is_flag=True, default=False,
              help='The flag to replace the existing deploy output file')
@click.option('--force', nargs=1, is_flag=True, default=False,
              help='The flag, to apply updates even if the latest deployment '
                   'failed')
@verbose_option
@check_deploy_name_for_duplicates
@check_deploy_bucket_exists
@check_bundle_deploy_names_for_existence()
def update(
        bundle_name: str,
        deploy_name: str,
        update_only_types: tuple | None = None,
        update_only_resources: tuple | None = None,
        update_only_resources_path: str | None = None,
        excluded_resources: tuple | None = None,
        excluded_resources_path: str | None = None,
        excluded_types: tuple | None = None,
        replace_output: bool = False,
        force: bool = False,
):
    """
    Updates infrastructure from the provided bundle
    """
    if not are_resource_types_valid(
            'excluded-types', excluded_types, UPDATE_RESOURCE_TYPE_PRIORITY) or \
            not are_resource_types_valid(
                'update-only-types', update_only_types, UPDATE_RESOURCE_TYPE_PRIORITY):
        return ABORTED_RETURN_CODE

    from syndicate.core import PROJECT_STATE
    USER_LOG.info(f'Bundle name: {bundle_name}')
    PROJECT_STATE.current_bundle = bundle_name

    if update_only_resources_path and os.path.exists(
            update_only_resources_path):
        update_resources_list = json.load(open(update_only_resources_path))
        update_only_resources = tuple(
            set(update_only_resources + tuple(update_resources_list)))
        if update_only_resources:
            USER_LOG.info(
                f'Resources to update: {list(update_only_resources)}')

    if excluded_resources_path and os.path.exists(excluded_resources_path):
        excluded_resources_list = json.load(open(excluded_resources_path))
        excluded_resources = tuple(
            set(excluded_resources + tuple(excluded_resources_list)))
        if excluded_resources:
            USER_LOG.info(
                f'Resources excluded from update: {list(excluded_resources)}')

    success = update_deployment_resources(
        bundle_name=bundle_name,
        deploy_name=deploy_name,
        update_only_types=update_only_types,
        update_only_resources=update_only_resources,
        excluded_resources=excluded_resources,
        excluded_types=excluded_types,
        replace_output=replace_output,
        force=force)
    if success is True:
        USER_LOG.info('Update of resources has been successfully completed')
        return OK_RETURN_CODE
    elif success == ABORTED_STATUS:
        USER_LOG.info('Update of resources has been aborted')
        return ABORTED_RETURN_CODE
    else:
        USER_LOG.warning('Something went wrong during resources update')
        return FAILED_RETURN_CODE


@syndicate.command(name=CLEAN_ACTION)
@return_code_manager
@sync_lock(lock_type=MODIFICATION_LOCK)
@timeit(action_name=CLEAN_ACTION)
@failed_status_code_on_exception
@click.option('--clean-only-types', '-types',
              cls=MultiWordOption, multiple=True, type=str,
              help='If specified only provided types will be cleaned')
@click.option('--clean-only-resources', '-resources',
              cls=MultiWordOption, multiple=True,
              help='If specified only provided resources will be cleaned')
@click.option('--clean-only-resources-path',
              '-path', cls=MultiWordOption, nargs=1, type=str,
              help='If specified only resources path will be cleaned')
@click.option('--clean-externals',
              cls=MultiWordOption, is_flag=True,
              help='Flag. If specified external resources will be '
                   'cleaned as well')
@click.option('--excluded-resources', '-exresources',
              cls=MultiWordOption, multiple=True,
              help='If specified provided resources will be excluded')
@click.option('--excluded-resources-path',
              '-expath', cls=MultiWordOption, nargs=1, type=str,
              help='If specified provided resource path will be excluded')
@click.option('--excluded-types', '-extypes',
              cls=MultiWordOption, multiple=True, type=str,
              help='If specified provided types will be excluded')
@click.option('--preserve-state',
              cls=MultiWordOption, is_flag=True,
              help='Preserve deploy output json file after resources removal')
@verbose_option
def clean(
        clean_only_types: tuple | None = None,
        clean_only_resources: tuple | None = None,
        clean_only_resources_path: str | None = None,
        clean_externals: bool = False,
        excluded_resources: tuple | None = None,
        excluded_resources_path: str | None = None,
        excluded_types: tuple | None = None,
        preserve_state: bool = False,
):
    """
    Cleans the application infrastructure
    """
    if not are_resource_types_valid(
            'excluded-types', excluded_types, CLEAN_RESOURCE_TYPE_PRIORITY) or \
            not are_resource_types_valid(
                'clean-only-types', clean_only_types, CLEAN_RESOURCE_TYPE_PRIORITY):
        return ABORTED_STATUS

    from syndicate.core import PROJECT_STATE
    if not PROJECT_STATE.latest_deploy:
        USER_LOG.error('Deployment to clean not found.')
        return ABORTED_RETURN_CODE

    bundle_name = PROJECT_STATE.latest_deployed_bundle_name
    deploy_name = PROJECT_STATE.latest_deployed_deploy_name
    USER_LOG.info('Command clean')
    USER_LOG.info(f'Deploy name: {deploy_name}')
    separator = ', '
    PROJECT_STATE.current_bundle = bundle_name

    if clean_only_types:
        USER_LOG.info(f'Clean only types: {separator.join(clean_only_types)}')
    if clean_only_resources:
        USER_LOG.info(f'Clean only resources: '
                      f'{separator.join(clean_only_resources)}')
    if clean_only_resources_path:
        USER_LOG.info(f'Clean only resources path: {clean_only_resources_path}')
    if excluded_resources:
        USER_LOG.info(f'Excluded resources: {separator.join(excluded_resources)}')
    if excluded_resources_path:
        USER_LOG.info(f'Excluded resources path: {excluded_resources_path}')
    if excluded_types:
        USER_LOG.info(f'Excluded types: {separator.join(excluded_resources)}')
    if clean_only_resources_path and os.path.exists(clean_only_resources_path):
        clean_resources_list = json.load(open(clean_only_resources_path))
        clean_only_resources = tuple(
            set(clean_only_resources + tuple(clean_resources_list)))
    if excluded_resources_path and os.path.exists(excluded_resources_path):
        excluded_resources_list = json.load(open(excluded_resources_path))
        excluded_resources = tuple(
            set(excluded_resources + tuple(excluded_resources_list)))

    result = remove_deployment_resources(
        deploy_name=deploy_name,
        bundle_name=bundle_name,
        clean_only_resources=clean_only_resources,
        clean_only_types=clean_only_types,
        excluded_resources=excluded_resources,
        excluded_types=excluded_types,
        clean_externals=clean_externals,
        preserve_state=preserve_state)

    if result == ABORTED_STATUS:
        USER_LOG.warning('Clean of resources has been aborted')
        return ABORTED_RETURN_CODE
    elif result is False:
        USER_LOG.warning('AWS resources were removed with errors.')
        return FAILED_RETURN_CODE
    elif isinstance(result, dict) and 'operation' in result:
        USER_LOG.info('AWS resources were removed.')
        return {**result, 'return_code': OK_RETURN_CODE}
    else:
        USER_LOG.info('AWS resources were removed.')
    return OK_RETURN_CODE


@syndicate.command(name=SYNC_ACTION)
@return_code_manager
@timeit()
@verbose_option
@check_deploy_bucket_exists
def sync():
    """
    Syncs the state of local project state file (.syndicate) and
    the remote one.
    """
    sync_project_state()
    return OK_RETURN_CODE


@syndicate.command(name=STATUS_ACTION, short_help='Shows the state of a local '
                                                  'project state file')
@return_code_manager
@click.option('--events', flag_value='events',
              callback=partial(validate_incompatible_options,
                               incompatible_options=['resources']),
              help='Show event logs of the project')
@click.option('--resources', flag_value='resources',
              callback=partial(validate_incompatible_options,
                               incompatible_options=['events']),
              help='Show a summary of the project resources')
@verbose_option
@timeit()
@check_deploy_bucket_exists
def status(events, resources):
    """
    Shows the state of a local project state file (.syndicate).
    Command displays the following content: project name, state, latest
    modification, locks summary, latest event, project resources.
    """
    click.echo(project_state_status(category=events or resources))
    return OK_RETURN_CODE


@syndicate.command(name=WARMUP_ACTION)
@return_code_manager
@sync_lock(lock_type=WARMUP_LOCK)
@timeit(action_name=WARMUP_ACTION)
@failed_status_code_on_exception
@click.option('--bundle-name', '-b', nargs=1,
              cls=MultiWordOption, callback=resolve_default_value,
              help='Name of the bundle. If not specified, resolves the latest '
                   'bundle name')
@click.option('--deploy-name', '-d', nargs=1,
              cls=MultiWordOption, callback=resolve_default_value,
              help='Name of the deploy. If not specified, resolves the latest '
                   'deploy name')
@click.option('--api-gw-id', '-api', nargs=1,
              cls=MultiWordOption, multiple=True, type=str,
              help='Provide API Gateway IDs to warmup')
@click.option('--stage-name', '-stage', nargs=1,
              cls=MultiWordOption, multiple=True, type=str,
              help='Name of stages of provided API Gateway IDs')
@click.option('--lambda-auth', '-auth',
              cls=MultiWordOption, default=False, is_flag=True,
              help='Flag. Should be specified if API Gateway Lambda Authorizer'
                   ' is enabled')
@click.option('--header-name', '-hname',
              cls=MultiWordOption, nargs=1,
              help='Name of authentication header.')
@click.option('--header-value', '-hvalue',
              cls=MultiWordOption, nargs=1,
              help='Authentication header value.')
@verbose_option
@check_deploy_bucket_exists
@check_bundle_deploy_names_for_existence()
def warmup(bundle_name, deploy_name, api_gw_id, stage_name, lambda_auth,
           header_name, header_value):
    """
    Warmups Lambda functions
    """
    if bundle_name and deploy_name:
        paths_to_be_triggered, resource_path_warmup_key_mapping = \
            process_deploy_resources(deploy_name=deploy_name,
                                     bundle_name=bundle_name)
    elif api_gw_id:
        paths_to_be_triggered, resource_path_warmup_key_mapping = \
            process_inputted_api_gw_id(api_id=api_gw_id, stage_name=stage_name)
    else:
        paths_to_be_triggered, resource_path_warmup_key_mapping = \
            process_existing_api_gw_id(stage_name=stage_name)

    if not paths_to_be_triggered or not resource_path_warmup_key_mapping:
        USER_LOG.info('No resources to warm up')
        return OK_RETURN_CODE
    resource_method_mapping, resource_warmup_key_mapping = \
        process_api_gw_resources(paths_to_be_triggered=paths_to_be_triggered,
                                 resource_path_warmup_key_mapping=
                                 resource_path_warmup_key_mapping)
    warm_upper(resource_method_mapping=resource_method_mapping,
               resource_warmup_key_mapping=resource_warmup_key_mapping,
               lambda_auth=lambda_auth, header_name=header_name,
               header_value=header_value)
    USER_LOG.info('Application resources have been warmed up.')
    return OK_RETURN_CODE


@syndicate.command(name=PROFILER_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b', nargs=1,
              cls=MultiWordOption, callback=resolve_default_value,
              help='The name of the bundle from which to select lambdas for '
                   'collecting metrics. If not specified, resolves the latest '
                   'bundle name')
@click.option('--deploy-name', '-d', nargs=1,
              cls=MultiWordOption, callback=resolve_default_value,
              help='Name of the deploy. If not specified, resolves the latest '
                   'deploy name')
@click.option('--from-date', '-from',
              cls=MultiWordOption, nargs=1,
              type=click.DateTime(formats=['%Y-%m-%dT%H:%M:%SZ']),
              help='Date from which collect lambda metrics. The '
                   '\'--to-date\' parameter required. Example of the date '
                   'format: 2022-02-02T02:02:02Z')
@click.option('--to-date', '-to',
              cls=MultiWordOption, nargs=1,
              type=click.DateTime(formats=['%Y-%m-%dT%H:%M:%SZ']),
              help='Date until which collect lambda metrics. The '
                   '\'--from-date\' parameter required. Example of the date '
                   'format: 2022-02-02T02:02:02Z')
@verbose_option
@check_deploy_bucket_exists
@check_bundle_deploy_names_for_existence(check_deploy_existence=True)
def profiler(bundle_name, deploy_name, from_date, to_date):
    """
    Displays application Lambda metrics
    """

    metric_value_dict = get_metric_statistics(bundle_name, deploy_name,
                                              from_date, to_date)
    for lambda_name, metrics in metric_value_dict.items():
        prettify_metrics_dict = {}

        USER_LOG.info(f'{os.linesep}Lambda function name: {lambda_name}')
        prettify_metrics_dict = process_metrics(prettify_metrics_dict, metrics)
        if not prettify_metrics_dict:
            USER_LOG.warning('No executions found')
        USER_LOG.info(tabulate(prettify_metrics_dict, headers='keys',
                               stralign='right'))
    return OK_RETURN_CODE


# =============================================================================


@syndicate.command(name=ASSEMBLE_JAVA_MVN_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b', cls=MultiWordOption, nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle, to which the build artifacts are '
                   'gathered and later used for the deployment. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--project-path', '-path', cls=MultiWordOption,
              nargs=1, callback=resolve_path_callback, required=True,
              help='The path to the Java project. The provided path is the '
                   'path for an mvn clean install. The artifacts are copied '
                   'to a folder, which is be later used as the deployment '
                   'bundle (the bundle path: bundles/${bundle_name})')
@click.option('--force-upload', '-F', cls=MultiWordOption,
              is_flag=True, default=False,
              help='Flag to override locally existing bundle '
                   'with the same name')
@click.option('--skip-tests', cls=MultiWordOption,
              is_flag=True, default=False,
              help='Flag to not run tests')
@click.option('--errors-allowed', cls=MultiWordOption,
              is_flag=True, default=False,
              help='Flag to continue building the bundle in case of errors '
                   'while building artifacts')
@verbose_option
@timeit(action_name=ASSEMBLE_JAVA_MVN_ACTION)
@failed_status_code_on_exception
def assemble_java_mvn(bundle_name, project_path, force_upload, skip_tests,
                      errors_allowed, refresh_cache=False, is_chained=False):
    """
    Builds Java lambdas

    \f
    :param bundle_name: name of the bundle
    :param project_path: path to project folder
    :param force_upload: force upload identification
    :param skip_tests: force skipping tests
    :param errors_allowed: not used for java, but need to unify the
    `assemble` commands interface
    :param refresh_cache: not used for java, but need to unify the function interface
    :param is_chained: specifies whether this command is executed independently
    or as part of a process (like `build` or `assemble` command)
    :return:
    """
    USER_LOG.info(f'Command compile java project path: {project_path}')

    if not is_chained:
        remove_bundle_dir_locally(bundle_name, force_upload)

    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_JAVA,
                       skip_tests=skip_tests,
                       errors_allowed=errors_allowed)
    USER_LOG.info('Java artifacts were prepared successfully.')
    return OK_RETURN_CODE


@syndicate.command(name=ASSEMBLE_PYTHON_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle, to which the build artifacts are '
                   'gathered and later used for the deployment. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--project-path', '-path',
              cls=MultiWordOption, nargs=1,
              callback=resolve_path_callback, required=True,
              help='The path to the Python project. The code is '
                   'packed to a zip archive, where the external libraries are '
                   'found, which are described in the requirements.txt file, '
                   'and internal project dependencies according to the '
                   'described in local_requirements.txt file')
@click.option('--force-upload', '-F',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to override locally existing bundle '
                   'with the same name')
@click.option('--errors-allowed',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to continue building the bundle if any errors occur '
                   'while building dependencies')
@click.option('--refresh-cache',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to refresh the cache of dependencies.')
@verbose_option
@timeit(action_name=ASSEMBLE_PYTHON_ACTION)
@failed_status_code_on_exception
def assemble_python(bundle_name, project_path, force_upload, errors_allowed,
                    refresh_cache, skip_tests=False, is_chained=False):
    """
    Builds Python lambdas

    \f
    :param bundle_name: name of the bundle
    :param project_path: path to project folder
    :param force_upload: force upload identification
    :param errors_allowed: allows to ignore dependency errors
    :param refresh_cache: flag to refresh the cache of dependencies
    :param skip_tests: not used for python, but need to unify the
    `assemble` commands interface
    :param is_chained: specifies whether this command is executed independently
    or as part of a process (like `build` or `assemble` command)
    :return:
    """
    USER_LOG.info(f'Command assemble python: project_path: {project_path} ')

    if not is_chained:
        remove_bundle_dir_locally(bundle_name, force_upload)

    if refresh_cache:
        cache_dir = resolve_bundles_cache_directory()
        if os.path.exists(cache_dir):
            shutil.rmtree(cache_dir)
            _LOG.info(f'Removed the cache directory: {cache_dir}')


    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_PYTHON,
                       errors_allowed=errors_allowed)
    USER_LOG.info('Python artifacts were prepared successfully.')
    return OK_RETURN_CODE


@syndicate.command(name=ASSEMBLE_NODE_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle, to which the build artifacts are '
                   'gathered and later used for the deployment. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--project-path', '-path',
              cls=MultiWordOption, nargs=1,
              callback=resolve_path_callback, required=True,
              help='The path to the NodeJS project. The code is '
                   'packed to a zip archive, where the external libraries are '
                   'found, which are described in the package.json file')
@click.option('--force-upload', '-F',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to override locally existing bundle '
                   'with the same name')
@verbose_option
@timeit(action_name=ASSEMBLE_NODE_ACTION)
@failed_status_code_on_exception
def assemble_node(bundle_name, project_path, force_upload,
                  errors_allowed=False, skip_tests=False, refresh_cache=False,
                  is_chained=False):
    """
    Builds NodeJS lambdas

    \f
    :param bundle_name: name of the bundle
    :param project_path: path to project folder
    :param force_upload: force upload identification
    :param errors_allowed: not used for NodeJS, but need to unify the
    `assemble` commands interface
    :param skip_tests: not used for NodeJS, but need to unify the
    `assemble` commands interface
    :param refresh_cache: not used for NodeJS, but need to unify the function interface
    :param is_chained: specifies whether this command is executed independently
    or as part of a process (like `build` or `assemble` command)
    :return:
    """
    USER_LOG.info(f'Command assemble node: project_path: {project_path} ')

    if not is_chained:
        remove_bundle_dir_locally(bundle_name, force_upload)

    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_NODEJS)
    USER_LOG.info('NodeJS artifacts were prepared successfully.')
    return OK_RETURN_CODE


@syndicate.command(name=ASSEMBLE_DOTNET_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle, to which the build artifacts are '
                   'gathered and later used for the deployment. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--project-path', '-path',
              cls=MultiWordOption, nargs=1,
              callback=resolve_path_callback, required=True,
              help='The path to the dotnet project. The code is '
                   'packed to a zip archive, where the external libraries are '
                   'found, which are described in the package.json file')
@click.option('--force-upload', '-F',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to override locally existing bundle '
                   'with the same name')
@verbose_option
@timeit(action_name=ASSEMBLE_DOTNET_ACTION)
@failed_status_code_on_exception
def assemble_dotnet(bundle_name, project_path, force_upload,
                    errors_allowed=False, skip_tests=False,
                    refresh_cache=False, is_chained=False):
    """
    Builds DotNet lambdas

    \f
    :param bundle_name: name of the bundle
    :param project_path: path to project folder
    :param force_upload: force upload identification
    :param errors_allowed: not used for DotNet, but need to unify the
    `assemble` commands interface
    :param skip_tests: not used for DotNet, but need to unify the
    `assemble` commands interface
    :param refresh_cache: not used for DotNet, but need to unify the function interface
    :param is_chained: specifies whether this command is executed independently
    or as part of a process
    :return:
    """
    USER_LOG.info(f'Command assemble dotnet: project_path: {project_path} ')

    if not is_chained:
        remove_bundle_dir_locally(bundle_name, force_upload)

    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_DOTNET)
    USER_LOG.info('DotNet artifacts were prepared successfully.')
    return OK_RETURN_CODE


@syndicate.command(name=ASSEMBLE_SWAGGER_UI_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle, to which the build artifacts are '
                   'gathered and later used for the deployment. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--project-path', '-path',
              cls=MultiWordOption, nargs=1,
              callback=resolve_path_callback, required=True,
              help='The path to the root project directory. Related files '
                   'will be packed into a zip archive.')
@click.option('--force-upload', '-F',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to override locally existing bundle '
                   'with the same name')
@verbose_option
@timeit(action_name=ASSEMBLE_SWAGGER_UI_ACTION)
@failed_status_code_on_exception
def assemble_swagger_ui(**kwargs):
    """
        Builds Swagger UI artifacts

        \f
        :param bundle_name: name of the bundle
        :param project_path: path to project folder
        :return:
        """
    bundle_name = kwargs.get('bundle_name')
    project_path = kwargs.get('project_path')
    USER_LOG.info(f'Command assemble Swagger UI: project-path: {project_path}')

    if not kwargs.get('is_chained'):
        remove_bundle_dir_locally(bundle_name, kwargs.get('force_upload'))

    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_SWAGGER_UI)
    USER_LOG.info('Swagger UI artifacts were prepared successfully.')
    return OK_RETURN_CODE


@syndicate.command(name=ASSEMBLE_APPSYNC_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, nargs=1,
              callback=generate_default_bundle_name,
              help='Name of the bundle, to which the build artifacts are '
                   'gathered and later used for the deployment. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--project-path', '-path',
              cls=MultiWordOption, nargs=1,
              callback=resolve_path_callback, required=True,
              help='Path to the project root directory. Related files '
                   'will be packed into a zip archive.')
@click.option('--force-upload', '-F',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to override locally existing bundle '
                   'with the same name')
@verbose_option
@timeit(action_name=ASSEMBLE_APPSYNC_ACTION)
@failed_status_code_on_exception
def assemble_appsync(**kwargs):
    """
        Builds AppSync artifacts

        \f
        :param bundle_name: name of the bundle
        :param project_path: path to project folder
        :return:
        """
    bundle_name = kwargs.get('bundle_name')
    project_path = kwargs.get('project_path')
    USER_LOG.info(f'Command assemble AppSync: project-path: {project_path} ')

    if not kwargs.get('is_chained'):
        remove_bundle_dir_locally(bundle_name, kwargs.get('force_upload'))

    assemble_artifacts(bundle_name=bundle_name,
                       project_path=project_path,
                       runtime=RUNTIME_APPSYNC)
    USER_LOG.info('AppSync artifacts were prepared successfully.')
    return OK_RETURN_CODE


RUNTIME_LANG_TO_BUILD_MAPPING = {
    JAVA_LANGUAGE_NAME: assemble_java_mvn,
    PYTHON_LANGUAGE_NAME: assemble_python,
    NODEJS_LANGUAGE_NAME: assemble_node,
    DOTNET_LANGUAGE_NAME: assemble_dotnet,
    SWAGGER_UI_NAME: assemble_swagger_ui,
    APPSYNC_NAME: assemble_appsync
}


@syndicate.command(name=ASSEMBLE_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b',
              cls=MultiWordOption,
              callback=generate_default_bundle_name,
              help='Bundle\'s name to build the lambdas in. '
                   'Default value: $ProjectName_%Y%m%d.%H%M%S')
@click.option('--force-upload', '-F',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to override locally existing bundle '
                   'with the same name')
@click.option('--errors-allowed',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag to continue building the bundle if any errors occur '
                   'while building dependencies. Only for Python runtime.')
@verbose_option
@click.pass_context
@timeit(action_name=ASSEMBLE_ACTION)
@failed_status_code_on_exception
def assemble(ctx, bundle_name, force_upload, errors_allowed, skip_tests=False,
             refresh_cache=False, is_chained=False):
    """
    Builds the application artifacts

    \f
    :param ctx:
    :param bundle_name: name of the bundle to which the artifacts
        will be associated
    :param force_upload: force upload identification
    :param errors_allowed: allows to ignore errors.
    :param skip_tests: allows to skip tests
    :param refresh_cache: specifies whether to refresh the cache of the syndicate
    :param is_chained: specifies whether this command is executed independently
    or as part of a process
    :return:
    """
    if not is_chained:
        remove_bundle_dir_locally(bundle_name, force_upload)

    USER_LOG.info(f'Building artifacts, bundle: {bundle_name}')
    from syndicate.core import PROJECT_STATE
    # Updates stale state aggregation
    PROJECT_STATE.refresh_state()

    build_mapping_dict: dict = PROJECT_STATE.load_project_build_mapping()
    if build_mapping_dict:
        for key, value in build_mapping_dict.items():
            func = RUNTIME_LANG_TO_BUILD_MAPPING.get(key)
            if func:
                return_code = ctx.invoke(func, bundle_name=bundle_name,
                                         project_path=value,
                                         errors_allowed=errors_allowed,
                                         skip_tests=skip_tests,
                                         refresh_cache=refresh_cache,
                                         is_chained=True)
                if return_code != OK_RETURN_CODE:
                    return return_code
            else:
                USER_LOG.error(f'Build tool is not supported: {key}')
                return FAILED_RETURN_CODE
    else:
        USER_LOG.info(
            'Resources for which artifacts need to be built were not found'
        )
    return OK_RETURN_CODE


@syndicate.command(name=PACKAGE_META_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, required=True,
              callback=verify_bundle_callback,
              help='Bundle\'s name to package the current meta in')
@verbose_option
@timeit(action_name=PACKAGE_META_ACTION)
@failed_status_code_on_exception
def package_meta(bundle_name):
    """
    Generates metadata about the application infrastructure

    \f
    :param bundle_name: name of the bundle to generate metadata
    :return:
    """
    from syndicate.core import CONFIG
    USER_LOG.info(f'Package meta, bundle: {bundle_name}')
    create_meta(project_path=CONFIG.project_path,
                bundle_name=bundle_name)
    USER_LOG.info('Meta was configured successfully.')
    return OK_RETURN_CODE


@syndicate.command(name=CREATE_DEPLOY_TARGET_BUCKET_ACTION)
@return_code_manager
@verbose_option
@timeit()
def create_deploy_target_bucket():
    """
    Creates a bucket in AWS account where all bundles will be uploaded
    """
    from syndicate.core import CONFIG
    USER_LOG.info(f'Create deploy target bucket: {CONFIG.deploy_target_bucket}')
    result = create_bundles_bucket()
    if not result:
        return ABORTED_RETURN_CODE
    USER_LOG.info('Deploy target bucket was created successfully')
    return OK_RETURN_CODE


@syndicate.command(name=UPLOAD_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b',
              cls=MultiWordOption,
              callback=resolve_and_verify_bundle_callback,
              help='Bundle name to which the build artifacts are gathered '
                   'and later used for the deployment. NOTE: if not '
                   'specified, the latest build will be uploaded')
@click.option('--force-upload','-F',
              cls=MultiWordOption, is_flag=True,
              help='Flag to override existing bundle with the same name as '
                   'provided')
@verbose_option
@timeit(action_name=UPLOAD_ACTION)
@failed_status_code_on_exception
@check_deploy_bucket_exists
def upload(bundle_name, force_upload=False):
    """
    Uploads bundle from local storage to AWS S3

    \f
    :param bundle_name: name of the bundle to upload
    :param force_upload: used if the bundle with the same name as provided
        already exists in an account
    :return:
    """
    USER_LOG.info(f'Upload bundle: {bundle_name}')
    if force_upload:
        USER_LOG.info('Force upload')

    futures = upload_bundle_to_s3(bundle_name=bundle_name, force=force_upload)
    handle_futures_progress_bar(futures)

    USER_LOG.info('Bundle was uploaded successfully')
    return OK_RETURN_CODE


@syndicate.command(name=COPY_BUNDLE_ACTION)
@return_code_manager
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, nargs=1,
              callback=create_bundle_callback, required=True,
              help='The bundle name, to which the build artifacts '
                   'are gathered and later used for the deployment')
@click.option('--src-account-id', '-acc-id',
              cls=MultiWordOption, nargs=1, required=True,
              help='The account ID, to which the bundle is to be '
                   'uploaded')
@click.option('--src-bucket-region', '-r',
              cls=MultiWordOption, nargs=1, required=True,
              type=ValidRegionParamType(),
              help='The name of the region of the bucket where target bundle '
                   'is stored')
@click.option('--src-bucket-name', '-bucket',
              cls=MultiWordOption, nargs=1, required=True,
              help='The name of the bucket where target bundle is stored')
@click.option('--role-name', '-role',
              cls=MultiWordOption, nargs=1, required=True,
              help='The role name from the specified account, which is '
                   'assumed. Here you have to check the trusted relationship '
                   'between the accounts. The active account must be a trusted'
                   ' one for the account which is specified in the command')
@click.option('--force-upload', '-F',
              cls=MultiWordOption, is_flag=True, default=False,
              help='Flag. Used if the bundle with the same name as provided '
                   'already exists in a target account')
@verbose_option
@timeit()
@click.pass_context
def copy_bundle(ctx, bundle_name, src_account_id, src_bucket_region,
                src_bucket_name, role_name, force_upload):
    """
    Copies the bundle from the specified account-region-bucket to
    account-region-bucket specified in syndicate.yml

    \f
    :param ctx:
    :param bundle_name: name of the bundle to copy
    :param src_account_id: id of the account where target bundle is stored
    :param src_bucket_region: region of the bucket where target bundle
        is stored
    :param src_bucket_name: name of the bucket where target bundle is stored
    :param role_name: name of the role that is assumed while copying
    :param force_upload: used if the bundle with the same name as provided
        already exists in a target account
    :return:
    """
    USER_LOG.info(f'Copy bundle: {bundle_name}')
    USER_LOG.info(f'Bundle name: {bundle_name}')
    USER_LOG.info(f'Source account id: {src_account_id}')
    USER_LOG.info(f'Source bucket region: {src_bucket_region}')
    USER_LOG.info(f'Source bucket name: {src_bucket_name}')
    futures = load_bundle(bundle_name, src_account_id, src_bucket_region,
                          src_bucket_name, role_name)
    handle_futures_progress_bar(futures)
    USER_LOG.info('Bundle was downloaded successfully')
    ctx.invoke(upload, bundle_name=bundle_name, force=force_upload)
    USER_LOG.info('Bundle was copied successfully')
    return OK_RETURN_CODE


@syndicate.command(name=EXPORT_ACTION)
@return_code_manager
@click.option('--resource-type', '-rt',
              cls=MultiWordOption, required=True,
              type=click.Choice(['api_gateway']),
              help='The type of resource to export configuration')
@click.option('--dsl', default='oas_v3',
              type=click.Choice(['oas_v3']),
              help='DSL of output specification. Default: oas_v3')
@click.option('--deploy-name', '-d', nargs=1,
              cls=MultiWordOption, callback=resolve_default_value,
              help='Name of the deploy. This parameter allows the framework '
                   'to decide, which exactly output file should be used. If '
                   'not specified, resolves the latest deploy name')
@click.option('--bundle-name', '-b',
              cls=MultiWordOption, callback=resolve_default_value,
              help='Name of the bundle to export from. Default value: name of '
                   'the latest built bundle')
@click.option('--output-dir', '-od', cls=MultiWordOption,
              help='The directory where an exported configuration will be '
                   'saved. If not specified, the directory with the name '
                   '"export" will be created in the project root directory to '
                   'store export files')
@verbose_option
def export(
        resource_type: str,
        dsl: str,
        deploy_name: str,
        bundle_name: str,
        output_dir: str | None = None,
):
    """
    Exports a configuration of the specified resource type to the file in a
    specified DSL

    param: resource_type: the type of the resource
    param: dsl: the DSL of the output configuration
    param: deploy_name: the name of the deployment
    param: bundle_name: the name of the bundle to export from
    param: output_dir: the directory where an exported specification will be
    saved
    """
    export_specification(
        resource_type=resource_type,
        dsl=dsl,
        deploy_name=deploy_name,
        bundle_name=bundle_name,
        output_directory=output_dir,
    )
    if resource_type == 'api_gateway' and dsl == 'oas_v3':
        USER_LOG.info(
            'Please note the AWS API Gateway-specific extensions are '
            'used to define the API in OAS v3 that starting with '
            '"x-amazon"'
        )
    return OK_RETURN_CODE


syndicate.add_command(generate)
syndicate.add_command(tags)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from boto3 import client
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry
from syndicate.core.helper import prettify_json

_LOG = get_logger(__name__)


@apply_methods_decorator(retry())
class RDSConnection(object):
    """ AWS RDS connection class. """

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.client = client('rds', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        self.region = region
        _LOG.debug('Opened new RDS connection.')

    def create_db_cluster(self, name: str, params: dict) -> dict:
        params = dict(
            DBClusterIdentifier=name,
            **params
        )
        _LOG.debug(f'DB cluster creation params: {prettify_json(params)}')

        return self.client.create_db_cluster(**params)['DBCluster']

    def create_db_instance(self, name: str, params: dict) -> dict:
        params = dict(
            DBInstanceIdentifier=name,
            **params
        )
        _LOG.debug(f'DB instance creation params: {prettify_json(params)}')

        return self.client.create_db_instance(**params)['DBInstance']

    def get_db_cluster_description(self, name: str) -> dict | None:
        try:
            response = self.client.describe_db_clusters(
                DBClusterIdentifier=name)
            return response['DBClusters'][0] if response['DBClusters'] \
                else {}
        except ClientError as e:

            if e.response['Error']['Code'] == 'DBClusterNotFoundFault':
                _LOG.info(f'RDS DB cluster {name} is not found')
                return {}

            else:
                raise e

    def get_db_instance_description(self, cluster_name: str = None,
                                    instance_name: str = None) -> dict:
        params = dict()

        if cluster_name:
            params['Filters'] = [
                {
                    'Name': 'db-cluster-id',
                    'Values': [cluster_name]
                }
            ]

        if instance_name:
            params['DBInstanceIdentifier'] = instance_name

        try:
            response = self.client.describe_db_instances(**params)
            return response['DBInstances'][0] if response['DBInstances'] \
                else {}
        except ClientError as e:

            if e.response['Error']['Code'] == 'DBInstanceNotFound':
                _LOG.info(f'RDS DB instance is not found')
                return {}

            else:
                raise e

    def get_db_cluster_endpoint(self, cluster_name: str) -> str:
        descr = self.get_db_cluster_description(cluster_name)
        return descr.get('Endpoint')

    def get_db_cluster_reader_endpoint(self, cluster_name: str) -> str:
        descr = self.get_db_cluster_description(cluster_name)
        return descr.get('ReaderEndpoint')

    def extract_secret_name_from_arn(self, arn: str) -> str:
        arn_tail = arn.split(':')[-1]
        name_parts = arn_tail.split('-')
        name_parts.pop(-1)
        return '-'.join(name_parts)

    def get_db_cluster_master_user_secret_name(self, cluster_name: str) -> str:
        descr = self.get_db_cluster_description(cluster_name)
        secret = descr.get('MasterUserSecret', {})
        if secret_arn := secret.get('SecretArn'):
            return self.extract_secret_name_from_arn(secret_arn)

    def get_waiter(self, waiter_name: str):
        return self.client.get_waiter(waiter_name)

    def update_db_instance(self, params: dict) -> dict:
        return self.client.modify_db_instance(**params)['DBInstance']

    def update_db_cluster(self, params: dict) -> dict:
        return self.client.modify_db_cluster(**params)['DBCluster']

    def delete_db_instance(self, instance_name: str):
        return self.client.delete_db_instance(
            DBInstanceIdentifier=instance_name,
            SkipFinalSnapshot=True,
            DeleteAutomatedBackups=True
        )

    def delete_db_cluster(self, name: str):
        return self.client.delete_db_cluster(
            DBClusterIdentifier=name,
            SkipFinalSnapshot=True
        )

    def is_db_cluster_del_protection(self, name: str) -> bool:
        description = self.get_db_cluster_description(name)
        return description.get('DeletionProtection')

    def is_db_instance_del_protection(self, name: str) -> bool:
        description = self.get_db_instance_description(name)
        return description.get('DeletionProtection')

    def disable_db_cluster_del_protection(self, name: str):
        _LOG.info(
            'Disabling DB cluster deletion protection...'
        )
        self.update_db_cluster({
            'DBClusterIdentifier': name,
            'DeletionProtection': False
        })

        _LOG.info(
            'Waiting for the DB cluster to become available...')
        waiter = self.get_waiter('db_cluster_available')
        waiter.wait(
            DBClusterIdentifier=name
        )

    def enable_db_cluster_del_protection(self, name: str):
        _LOG.info(
            'Enabling DB cluster deletion protection...'
        )
        self.update_db_cluster({
            'DBClusterIdentifier': name,
            'DeletionProtection': True
        })

        _LOG.info(
            'Waiting for the DB cluster to become available...')
        waiter = self.get_waiter('db_cluster_available')
        waiter.wait(
            DBClusterIdentifier=name
        )

    def disable_db_instance_del_protection(self, name: str):
        _LOG.info(
            'Disabling DB instance deletion protection...'
        )
        self.update_db_instance({
            'DBInstanceIdentifier': name,
            'DeletionProtection': False
        })

        _LOG.info(
            'Waiting for the DB instance to become available...')
        waiter = self.get_waiter('db_instance_available')
        waiter.wait(
            DBInstanceIdentifier=name
        )



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.exceptions import InvalidValueError, ParameterError, \
    InvalidTypeError
from syndicate.core.constants import OPTIMAL_INSTANCE_TYPE

COMPENV_STATES = ('ENABLED', 'DISABLED')
COMPENV_TYPES = ('UNMANAGED', 'MANAGED')
COMPUTE_RESOURCE_TYPES = ('EC2', 'SPOT', 'FARGATE', 'FARGATE_SPOT')
FARGATE_RESOURCE_TYPES = ('FARGATE', 'FARGATE_SPOT')
ALLOCATION_STRATEGIES = ('BEST_FIT', 'BEST_FIT_PROGRESSIVE', 'SPOT_CAPACITY_OPTIMIZED')


def _get_ec2_connection():
    from syndicate.core import CONN
    return CONN.ec2()

def validate_batch_compenv(compenv_name, compenv_meta):
    """
    Performs check of Batch Compute Environment resources.
    :param compenv_name: name of resource
    :param compenv_meta: resource definition

    :raises an exception in case of invalidity.

    :return: None
    """
    # Config stores a list of validation rules for each parameter.
    # Each rule contains:
    # - 'field name' - variable name that will be displayed in case of invalidity
    # - 'field value' - parameter value
    # - 'validators' - list of validator functions to test parameter
    # - any other arguments that will be passed to validator
    config = [
        {
            'field_name': 'compute_environment_name',
            'field_value': compenv_name,
            'prefix': '',
            'validators': [
                _validate_required_field
            ]
        },
        {
            'field_name': 'state',
            'field_value': compenv_meta.get('state'),
            'field_options': COMPENV_STATES,
            'prefix': '',
            'required': False,
            'validators': [
                _validate_options_field
            ]
        },
        {
            'field_name': 'compute_environment_type',
            'field_value': compenv_meta.get('compute_environment_type'),
            'field_options': COMPENV_TYPES,
            'prefix': '',
            'required': True,
            'validators': [
                _validate_options_field
            ]
        },
    ]
    _process_config(config)

    compute_resources = compenv_meta.get('compute_resources')
    if compute_resources:
        _validate_compute_resources(compute_resources)


def _validate_compute_resources(compute_resources):
    """
    Performs check of Batch Compute Environment compute resources.
    :param compute_resources: compute resources definition

    :raises an exception in case of invalidity.

    :return: None
    """
    ec2_conn = _get_ec2_connection()

    compute_resource_type = compute_resources.get('type')
    compute_resource_config = [
        {
            'field_name': 'compute_resource_type',
            'field_value': compute_resource_type,
            'field_options': COMPUTE_RESOURCE_TYPES,
            'required': True,
            'validators': [
                _validate_options_field
            ]
        },
        {
            'field_name': 'allocation_strategy',
            'field_value': compute_resources.get('allocation_strategy'),
            'field_options': ALLOCATION_STRATEGIES,
            'compute_resource_type': compute_resource_type,
            'required': False,
            'validators': [
                _validate_options_field,
                _validate_fargate_forbidden_field
            ]
        },
        {
            'field_name': 'min_vcpus',
            'field_value': compute_resources.get('minv_cpus'),
            'compute_resource_type': compute_resource_type,
            'required_type': int,
            'validators': [
                _validate_field_type,
                _validate_fargate_forbidden_field
            ]
        },
        {
            'field_name': 'maxv_cpus',
            'field_value': compute_resources.get('maxv_cpus'),
            'compute_resource_type': compute_resource_type,
            'required_type': int,
            'validators': [
                _validate_required_field,
                _validate_field_type,
            ]
        },
        {
            'field_name': 'desiredv_cpus',
            'field_value': compute_resources.get('desiredv_cpus'),
            'compute_resource_type': compute_resource_type,
            'required_type': int,
            'validators': [
                _validate_field_type,
                _validate_fargate_forbidden_field
            ]
        },
        {
            'field_name': 'instance_types',
            'field_value': compute_resources.get('instance_types'),
            'compute_resource_type': compute_resource_type,
            'required_type': list,
            'validators': [
                _validate_field_type,
                _validate_fargate_forbidden_field,
            ]
        },
        {
            'field_name': 'image_id',
            'field_value': compute_resources.get('image_id'),
            'compute_resource_type': compute_resource_type,
            'required_type': str,
            'validators': [
                _validate_field_type,
                _validate_fargate_forbidden_field
            ]
        },
        {
            'field_name': 'subnets',
            'field_value': compute_resources.get('subnets'),
            'required_type': list,
            'validators': [
                _validate_required_field,
                _validate_field_type,
            ]
        },
        {
            'field_name': 'security_group_ids',
            'field_value': compute_resources.get('security_group_ids'),
            'required_type': list,
            'validators': [
                _validate_field_type,
            ]
        },
        {
            'field_name': 'ec2_key_pair',
            'field_value': compute_resources.get('ec2_key_pair'),
            'required_type': str,
            'compute_resource_type': compute_resource_type,
            'validators': [
                _validate_field_type,
                _validate_fargate_forbidden_field
            ]
        },
        {
            'field_name': 'instance_role',
            'field_value': compute_resources.get('instance_role'),
            'required_type': str,
            'compute_resource_type': compute_resource_type,
            'validators': [
                _validate_field_type,
                _validate_fargate_forbidden_field
            ]
        },
        {
            'field_name': 'tags',
            'field_value': compute_resources.get('tags'),
            'required_type': dict,
            'compute_resource_type': compute_resource_type,
            'validators': [
                _validate_field_type,
                _validate_fargate_forbidden_field
            ]
        },
        {
            'field_name': 'placement_group',
            'field_value': compute_resources.get('placement_group'),
            'required_type': str,
            'compute_resource_type': compute_resource_type,
            'validators': [
                _validate_field_type,
                _validate_fargate_forbidden_field
            ]
        },
        {
            'field_name': 'bid_percentage',
            'field_value': compute_resources.get('bid_percentage'),
            'required_type': int,
            'compute_resource_type': compute_resource_type,
            'validators': [
                _validate_field_type,
                _validate_fargate_forbidden_field
            ]
        },
        {
            'field_name': 'spot_iam_fleet_role',
            'field_value': compute_resources.get('spot_iam_fleet_role'),
            'required_type': str,
            'compute_resource_type': compute_resource_type,
            'validators': [
                _validate_field_type,
                _validate_fargate_forbidden_field
            ]
        },
        {
            'field_name': 'launch_template',
            'field_value': compute_resources.get('launch_template'),
            'required_type': dict,
            'compute_resource_type': compute_resource_type,
            'validators': [
                _validate_field_type,
                _validate_fargate_forbidden_field
            ]
        },
        {
            'field_name': 'ec2_configuration',
            'field_value': compute_resources.get('ec2_configuration'),
            'required_type': list,
            'validators': [
                _validate_field_type,
            ]
        },
    ]
    _process_config(compute_resource_config)

    instance_types = compute_resources.get('instance_types') or []
    available = set(ec2_conn.get_instance_types())

    available.add(OPTIMAL_INSTANCE_TYPE)
    for instance_type in instance_types:
        _validate_options_field(
            field_name='instance_types__item',
            field_value=instance_type,
            field_options=available
        )

    desiredv_cpus = compute_resources.get('desiredv_cpus')
    if desiredv_cpus:
        minv_cpus = compute_resources.get('minv_cpus')
        maxv_cpus = compute_resources.get('maxv_cpus')
        if desiredv_cpus > maxv_cpus:
            raise InvalidValueError(
                "'compute_resources__desired_vcpus' must be smaller or equal "
                "than 'max_vcpus'."
            )
        if minv_cpus and desiredv_cpus < minv_cpus:
            raise InvalidValueError(
                "'compute_resources__desired_vcpus' must be greater or equal "
                "than 'min_vcpus'."
            )

    security_group_ids = compute_resources.get('security_group_ids')
    if not security_group_ids and compute_resource_type in FARGATE_RESOURCE_TYPES:
        raise ParameterError(
            "'compute_resources__security_group_ids' is required for jobs "
            "running on Fargate resources"
        )

    launch_template = compute_resources.get('launch_template')
    if launch_template:
        launch_template_id = launch_template.get('launch_template_id')
        _validate_field_type(
            field_name='launch_template_id',
            field_value=launch_template_id,
            prefix='compute_resources__launch_template',
            required_type=str
        )

        launch_template_name = launch_template.get('launch_template_name')
        _validate_field_type(
            field_name='launch_template_name',
            field_value=launch_template_name,
            prefix='compute_resources__launch_template',
            required_type=str
        )

        launch_template_options = (launch_template_id, launch_template_name)
        if all(launch_template_options) or not any(launch_template_options):
            raise ParameterError(
                "You must specify either the 'launch_template_id' or "
                "'launch_template_name', but not both."
            )


def _validate_options_field(field_name, field_value, field_options, prefix='compute_resources', required=True,
                            **kwargs):
    """
    Checks whether a field contains value from options list.

    :param field_name: name of field that will be displayed in case of invalidity
    :param field_value: value of field to check
    :param field_options: list of options that field can accept
    :param prefix: prefix that will be displayed before field_name in case of invalidity
    :param required: if field is required and can be empty

    :raises an exception in case of invalidity.

    :return: None
    """
    if prefix:
        field_name = prefix + '__' + field_name

    if not required and not field_value:
        return
    if required and not field_value:
        raise ParameterError(
            f"Missing required Compute Environment field: '{field_name}'"
        )
    if field_value not in field_options:
        raise InvalidValueError(
            f"Compute Environment field: '{field_name}': '{str(field_value)}' "
            f"must be one of the following: '{field_options}'"
        )


def _validate_fargate_forbidden_field(field_name, field_value, compute_resource_type, prefix='compute_resources',
                                      **kwargs):
    """
    Checks whether a field is forbidden to set for Fargate resources.

    :param field_name: name of field that will be displayed in case of invalidity
    :param field_value: value of field to check
    :param compute_resource_type: current compute resource type
    :param prefix: prefix that will be displayed before field_name in case of invalidity

    :raises an exception in case if compute resource type belongs to Fargate resources and given field is not empty.

    :return: None
    """
    if prefix:
        field_name = prefix + '__' + field_name
    if field_value and compute_resource_type in FARGATE_RESOURCE_TYPES:
        raise ParameterError(
            f"'{field_name}' parameter isn't applicable to jobs running on "
            f"Fargate resources, and shouldn't be specified."
        )


def _validate_required_field(field_name, field_value, prefix='compute_resources', **kwargs):
    """
    Checks whether a field is not empty.

    :param field_name: name of field that will be displayed in case of invalidity
    :param field_value: value of field to check
    :param prefix: prefix that will be displayed before field_name in case of invalidity

    :raises an exception in case of invalidity.

    :return: None
    """
    if prefix:
        field_name = prefix + '__' + field_name

    if not field_value:
        raise ParameterError(
            f"Missing required Compute Environment field: '{field_name}'"
        )


def _validate_field_type(field_name, field_value, required_type, prefix='compute_resources', **kwargs):
    """
    Checks whether a field is instance of the given type.

    :param field_name: name of field that will be displayed in case of invalidity
    :param field_value: value of field to check
    :param required_type: required type
    :param prefix: prefix that will be displayed before field_name in case of invalidity

    :raises an exception in case of invalidity.

    :return: None
    """
    if prefix:
        field_name = prefix + '__' + field_name

    if field_value is not None and not isinstance(field_value, required_type):
        raise InvalidTypeError(
            f"'{field_name}' parameter must be a '{required_type.__name__}'."
        )


def _process_config(config):
    for parameter in config:
        for validator in parameter.get('validators'):
            validator(**parameter)



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from syndicate.exceptions import ParameterError


def assert_required_property(resource_name, property_name, property_value,
                             error_msg=None):
    if property_value is None:
        error_msg = error_msg if error_msg else (
            f"Property '{property_name}' of resource '{resource_name}' is "
            f"required but absent"
        )
        raise ParameterError(error_msg)



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
import os

import click

from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.build.helper import resolve_bundle_directory
from syndicate.core.constants import BUILD_META_FILE_NAME
from syndicate.core.helper import build_path
from syndicate.core.transform.cloudformation.cloudformation_transformer import \
    CloudFormationTransformer
from syndicate.core.transform.terraform.terraform_transformer import TerraformTransformer

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()

TERRAFORM_DSL = 'terraform'
CLOUD_FORMATION_DSL = 'cloudformation'

TRANSFORM_PROCESSORS = {
    TERRAFORM_DSL: TerraformTransformer,
    CLOUD_FORMATION_DSL: CloudFormationTransformer
}


def generate_build_meta(bundle_name, dsl_list, output_directory):
    bundle_path = resolve_bundle_directory(bundle_name=bundle_name)
    build_meta_path = build_path(bundle_path, BUILD_META_FILE_NAME)
    with open(build_meta_path) as build_meta_file:
        meta_resources = json.load(build_meta_file)

    for dsl in dsl_list:

        transformer_type = TRANSFORM_PROCESSORS.get(dsl)
        if not transformer_type:
            USER_LOG.warning(f'There is no transformer for the dsl: {dsl}')
            continue

        transformer = transformer_type(bundle_name=bundle_name)
        transformed_build_meta = \
            transformer.transform_build_meta(build_meta=meta_resources)
        if not output_directory:
            output_path = build_path(
                bundle_path, transformer.output_file_name())
        elif not os.path.isabs(output_directory):
            output_path = build_path(
                os.getcwd(), output_directory, transformer.output_file_name())
        else:
            output_path = build_path(
                output_directory, transformer.output_file_name())
        with open(output_path, 'w') as output_file:
            output_file.write(transformed_build_meta)
        USER_LOG.info(
            f"The '{bundle_name}' bundle is transformed into {dsl} dsl."
        )



"""
    Copyright 2021 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
from datetime import datetime

from tabulate import tabulate

from syndicate.core.project_state.project_state import (
    OPERATION_LOCK_MAPPINGS, MODIFICATION_LOCK, WARMUP_LOCK,
    LOCK_LAST_MODIFICATION_DATE, LOCK_LOCKED_TILL)
from syndicate.core.project_state.sync_processor import sync_project_state
from syndicate.core.constants import DATE_FORMAT_ISO_8601, MODIFICATION_OPS

LOCKS = {
    MODIFICATION_LOCK: 'modification',
    WARMUP_LOCK: 'warmup'
}

LINE_SEP = os.linesep


def project_state_status(category=None):
    sync_project_state()
    if category == 'events':
        return process_events_view()
    elif category == 'resources':
        return process_resources_view()
    else:
        return process_default_view()


def process_default_view():
    from syndicate.core import PROJECT_STATE
    project_name = PROJECT_STATE.name
    events = PROJECT_STATE.events
    last_event = None
    latest_operation = None
    is_locked = False
    if events:
        last_event = events[0]
        latest_operation = last_event.get('operation')
        lock_name = OPERATION_LOCK_MAPPINGS.get(latest_operation)
        is_locked = not PROJECT_STATE.is_lock_free(lock_name)

    state = 'Locked' if is_locked else 'Available'

    result = [
        f'Project: {project_name}',
        f'State: {state}',
        LINE_SEP + locks_summary(PROJECT_STATE)
    ]

    last_modification = PROJECT_STATE.latest_modification
    if last_modification and latest_operation not in MODIFICATION_OPS:
        operation = last_modification.get('operation')
        result.append(LINE_SEP + f'Latest modification: {operation}')
        modification_start_time = last_modification.get('time_start')
        modification_end_time = last_modification.get('time_end')
        data = [
            ['', 'Bundle name: ', last_modification.get('bundle_name')],
            ['', 'Deploy name: ', last_modification.get('deploy_name')],
            ['', 'Initiated by: ', last_modification.get('initiator')],
            ['', 'Started at: ', format_time(modification_start_time)],
            ['', 'Ended at: ', format_time(modification_end_time)],
            ['', 'Duration (sec): ', f"{last_modification.get('duration_sec')}"],
            ['', 'Status: ', f"{last_modification.get('status')}"],
        ]
        result.append(tabulate_data(data))

    if last_event:
        latest_operation = last_event.get('operation')
        result.append(LINE_SEP + f'Latest event: {latest_operation}')
        event_start_time = last_event.get('time_start')
        event_end_time = last_event.get('time_end')
        data = [
            ['', 'Bundle name: ', last_event.get('bundle_name')],
            ['', 'Initiated by: ', last_event.get('initiator')],
            ['', 'Started at: ', format_time(event_start_time)],
            ['', 'Ended at: ', format_time(event_end_time)],
            ['', 'Duration (sec): ', f"{last_event.get('duration_sec')}"],
            ['', 'Status: ', f"{last_event.get('status')}"],
        ]
        if last_event.get('deploy_name'):
            data.insert(1,
                        ['', 'Deploy name: ', last_event.get('deploy_name')])
        result.append(tabulate_data(data))
    lambdas = PROJECT_STATE.lambdas
    result.append(LINE_SEP + 'Project resources:')
    if lambdas:
        headers = ['Type', 'Quantity']
        resources = [['Lambda', len(lambdas)]]
        result.append(
            tabulate_data(data=resources, headers=headers, tablefmt='simple')
        )
    else:
        result.append(indent('There are no lambdas in this project.'))
    return LINE_SEP + LINE_SEP.join(result)


def process_events_view():
    from syndicate.core import PROJECT_STATE
    project_name = PROJECT_STATE.name
    result = ['Project: {}'.format(project_name),
              'Event logs:']
    events = PROJECT_STATE.events
    if events:
        headers = ['Operation', 'Started at', 'Duration (sec)',
                   'Initiator', 'Bundle', 'Deploy', 'Status']
        summaries = []
        for event in events:
            summaries.append([
                event.get('operation'),
                format_time(event.get('time_start')),
                event.get('duration_sec'),
                event.get('initiator'),
                event.get('bundle_name'),
                event.get('deploy_name'),
                event.get('status')
            ])
        result.append(tabulate_data(data=summaries, headers=headers,
                                    tablefmt='simple'))
    else:
        result.append('There are no events regarding this project.')
    return LINE_SEP + LINE_SEP.join(result)


def process_resources_view():
    from syndicate.core import PROJECT_STATE
    project_name = PROJECT_STATE.name
    result = ['Project: {}'.format(project_name),
              'Lambda resources:']
    lambdas = PROJECT_STATE.lambdas
    if lambdas:
        headers = ['Name', 'Runtime']
        lambdas_data = []
        for lambda_name, lambda_info in lambdas.items():
            lambdas_data.append([lambda_name, lambda_info.get('runtime')])
        result.append(tabulate_data(data=lambdas_data, headers=headers,
                                    tablefmt='simple'))
    else:
        result.append(indent('There are no lambdas in this project.'))
    return LINE_SEP + LINE_SEP.join(result)


def locks_summary(project_state):
    result = ['Locks summary:']
    current_locks = project_state.locks
    all_locks = {
        MODIFICATION_LOCK: current_locks.get(MODIFICATION_LOCK),
        WARMUP_LOCK: current_locks.get(WARMUP_LOCK)
    }
    headers = ['Type', 'State', 'Last modification date', 'Locked till']
    locks_data = []
    for lock_name, lock_info in all_locks.items():
        display_name = LOCKS.get(lock_name)
        if lock_info:
            state = 'Acquired' if lock_info.get(LOCK_LOCKED_TILL) else 'Released'
            last_mod_date = lock_info.get(LOCK_LAST_MODIFICATION_DATE)
            last_mod_date = format_time(last_mod_date)
            locked_till_date = lock_info.get(LOCK_LOCKED_TILL)
            if locked_till_date:
                locked_till_date = format_time(locked_till_date)
            locks_data.append([
                display_name, state, last_mod_date, locked_till_date])
        else:
            locks_data.append([display_name, 'Released', None, None])
    result.append(tabulate_data(data=locks_data, headers=headers,
                                tablefmt='simple'))
    return LINE_SEP.join(result)


def tabulate_data(data, headers=(), tablefmt='plain'):
    return tabulate(tabular_data=data, tablefmt=tablefmt,
                    floatfmt='.3f', headers=headers, missingval='-')


def indent(line):
    return '    ' + line


def format_time(time_string):
    time = datetime.strptime(time_string, DATE_FORMAT_ISO_8601)
    return time.strftime('%Y-%m-%d %H:%M:%S')



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import json
from secrets import token_hex
from typing import Optional

from boto3 import client
from botocore.client import BaseClient
from botocore.exceptions import ClientError

from syndicate.commons.log_helper import get_logger
from syndicate.connection.helper import apply_methods_decorator, retry

RESPONSE_PARAM_ALLOW_ORIGIN = \
    "method.response.header.Access-Control-Allow-Origin"
RESPONSE_PARAM_ALLOW_METHODS = \
    "method.response.header.Access-Control-Allow-Methods"
RESPONSE_PARAM_ALLOW_HEADERS = \
    "method.response.header.Access-Control-Allow-Headers"

REQ_VALIDATOR_PARAM_VALIDATE_PARAMS = 'validateRequestParameters'
REQ_VALIDATOR_PARAM_VALIDATE_BODY = 'validateRequestBody'
REQ_VALIDATOR_PARAM_NAME = 'name'

_LOG = get_logger(__name__)


@apply_methods_decorator(retry())
class ApiGatewayConnection(object):
    """ API Gateway connection class."""

    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self.region = region
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.client = client('apigateway', region,
                             aws_access_key_id=aws_access_key_id,
                             aws_secret_access_key=aws_secret_access_key,
                             aws_session_token=aws_session_token)
        _LOG.debug('Opened new API Gateway connection.')

    def create_rest_api(self, api_name,
                        binary_media_types=None,
                        description=None,
                        clone_from=None,
                        tags=None):
        """
        :type api_name: str
        :type description: str
        :type binary_media_types: list
        :type clone_from: str
        :param clone_from: The ID of the RestApi that you want to clone from.
        :param tags: dict The resource tags key-value pairs
        """
        params = dict(name=api_name)
        if description:
            params['description'] = description
        if clone_from:
            params['cloneFrom'] = clone_from
        if binary_media_types:
            params['binaryMediaTypes'] = binary_media_types
        if tags:
            params['tags'] = tags
        return self.client.create_rest_api(**params)

    def create_openapi(self, openapi_context):
        # Create a new API Gateway with the OpenAPI definition
        response = self.client.import_rest_api(
            body=json.dumps(openapi_context),
            failOnWarnings=False
        )
        api_id = response['id']
        _LOG.debug(f"API Gateway created successfully with ID: {api_id}")
        return api_id

    def tag_openapi(self, openapi_id, tags):
        try:
            self.client.tag_resource(
                resourceArn=\
                    f'arn:aws:apigateway:{self.client.meta.region_name}'
                    f'::/restapis/{openapi_id}',
                tags={**tags}
            )
            _LOG.debug(f"Tags: {tags} applied to openapi with id:"
                       f" {openapi_id}.")
        except self.client.exceptions.BadRequestException as e:
            _LOG.error(f"Unexpected error happened while tagging openapi: {e}")

    def describe_openapi(self, api_id, stage_name):
        try:
            response = self.client.get_export(
                restApiId=api_id,
                stageName=stage_name,
                exportType='oas30',
                parameters={
                    'extensions': 'integrations,authorizers,apigateway'},
                accepts='application/json'
            )
            return response
        except self.client.exceptions.NotFoundException:
            _LOG.warning(f"Not found api with id: {api_id} "
                         f"and stage name: {stage_name}")
            return None

    def describe_tags(self, api_arn: str) -> dict | None:
        try:
            response = self.client.get_tags(resourceArn=api_arn)
            return response
        except self.client.exceptions.BadRequestException as e:
            _LOG.error(f"Failed to retrieve tags for ARN {api_arn}: {str(e)}")
            return None

    def update_openapi(self, api_id, openapi_context):
        # Update the API Gateway with the OpenAPI definition
        self.client.put_rest_api(
            restApiId=api_id,
            mode='overwrite',
            body=json.dumps(openapi_context),
            failOnWarnings=False
        )
        _LOG.debug("API Gateway updated successfully.")

    def remove_api(self, api_id, log_not_found_error=True):
        """
        :type api_id: str
        :type log_not_found_error: boolean, parameter is needed for proper
        log handling in the retry decorator
        """
        self.client.delete_rest_api(restApiId=api_id)

    def get_api_by_name(self, api_name):
        """
        :type api_name: str
        """
        apis = self.get_all_apis()
        target_apis = [api for api in apis if api['name'] == api_name]
        if len(target_apis) == 1:
            return target_apis[0]
        if len(target_apis) > 1:
            _LOG.warn(f"API Gateway can\'t be identified unambiguously "
                      f"because there is more than one resource with the name "
                      f"'{api_name}' in the region {self.region}. Determined "
                      f"APIs: {[api['name'] for api in target_apis]}")
        else:
            _LOG.warn(f'API Gateway with the name "{api_name}" '
                      f'not found in the region {self.region}')

    def get_api_id(self, api_name):
        """
        :type api_name: str
        """
        api_info = self.get_api_by_name(api_name)
        if api_info:
            return api_info['id']

    def get_resource_id(self, api_id, resource_path):
        """
        :type api_id: str
        :type resource_path: str
        """
        resource_info = self.get_resource_by_path(api_id, resource_path)
        if resource_info:
            return resource_info['id']

    def get_resource_by_path(self, api_id, resource_path):
        """
        :type api_id: str
        :type resource_path: str
        """
        if not api_id:
            return
        resources = []
        response = self.client.get_resources(restApiId=api_id, limit=100)
        if response.get('items'):
            resources.extend(response.get('items'))
        token = response.get('position')
        while token:
            response = self.client.get_resources(restApiId=api_id, limit=100,
                                                 position=token)
            if response.get('items'):
                resources.extend(response.get('items'))
            token = response.get('position')
        for each in resources:
            if each['path'] == resource_path:
                return each

    def get_resources(self, api_id):
        resources = []
        response = self.client.get_resources(restApiId=api_id)
        if response.get('items'):
            resources.extend(response.get('items'))
        token = response.get('position')
        while token:
            response = self.client.get_resources(restApiId=api_id,
                                                 position=token)
            if response.get('items'):
                resources.extend(response.get('items'))
            token = response.get('position')
        return resources

    def get_method(self, api_id, resource_id, method):
        """
        :type api_id: str
        :type resource_id: str
        :type method: str
        """
        res = self.client.get_resource(restApiId=api_id,
                                       resourceId=resource_id)
        methods = res.get('resourceMethods')
        if methods and method in methods:
            return True

    def create_resource(self, api_id, resource_path):
        """
        :type api_id: str
        :type resource_path: str
        """
        initial_resource = '/'
        parent_resource_id = self.get_resource_id(api_id, initial_resource)
        _LOG.debug('Processing resource %s, parent id - %s', initial_resource,
                   parent_resource_id)
        if resource_path.startswith('/'):
            resource_path = resource_path[1:]
        resource_path_split = resource_path.split('/')
        for resource in resource_path_split:
            try:
                _LOG.debug('Processing parent id: %s', parent_resource_id)
                resp = self.client.create_resource(restApiId=api_id,
                                                   parentId=parent_resource_id,
                                                   pathPart=resource)
                resource_id = resp['id']
            except ClientError as e:
                resource_id = None
                if 'ConflictException' in str(e):
                    _LOG.debug('Error while creating resource {0}.'
                               ' Creation stage: {1}.'
                               .format(resource_path, resource))
                    resource_info = self.get_resource_by_path(
                        api_id, initial_resource + resource)
                    if resource_info:
                        resource_id = resource_info['id']
                if not resource_id:
                    raise e
            parent_resource_id = resource_id
            initial_resource += resource
            initial_resource += '/'

    def create_method(self, api_id, resource_id, method,
                      authorization_type=None, authorizer_id=None,
                      api_key_required=None, request_parameters=None,
                      request_models=None,
                      request_validator=None):
        """
        :type api_id: str
        :type resource_id: str
        :type method: str
        :param method: Specifies the method request's HTTP method type.
        :type authorization_type: str
        :param authorization_type: Specifies the type of authorization used for
        the method.
        :type authorizer_id: str
        :param authorizer_id: Specifies the identifier of an Authorizer to use
        on this Method, if the type is CUSTOM.
        :type api_key_required: bool
        :param api_key_required: Specifies whether the method required a valid
        ApiKey .
        :type request_parameters: dict
        :param request_parameters: A key-value map defining required or
        optional method request parameters that can be accepted by API Gateway.
        {'string': True|False}. Boolean flag - required parameter or optional.
        :type request_models: dict
        :param request_models: Specifies the Model resources used for the
        request's content type. Request models are represented as a key/value
        map, with a content type as the key and a Model name as the value.
        {'string': 'string'}
        :type request_validator: dict
        :param request_validator: Dictionary with values to create request
        validator. It could contains name of validator and validate parameters
        (validate_request_body, validate_request_parameters or both).
        """
        params = dict(restApiId=api_id, resourceId=resource_id,
                      httpMethod=method, authorizationType='NONE',
                      apiKeyRequired=False)
        if authorization_type:
            params['authorizationType'] = authorization_type
        if authorizer_id:
            params['authorizerId'] = authorizer_id
        if api_key_required:
            params['apiKeyRequired'] = api_key_required
        if request_parameters:
            params['requestParameters'] = request_parameters
        if request_models:
            params['requestModels'] = request_models
        if request_validator:
            params['requestValidatorId'] = request_validator
        self.client.put_method(**params)

    def create_request_validator(self, api_id, name: str = None,
                                 validate_request_body: bool = False,
                                 validate_request_parameters: bool = False):
        """
        Helper function to create a request validator. Returns its id
        :type api_id: str
        :param api_id: Identifier of the associated RestApi.
        :type name: str
        :param name: name of request validator. If not set, will be generated
        :type validate_request_body: bool = False
        :param validate_request_body: whether to validate request body
        :type validate_request_parameters: bool = False
        :param validate_request_parameters: whether to validate query params
        :return: str, identifier of created RequestValidator.
        """
        params = dict(
            restApiId=api_id,
            name=name or f'default-validator-name-{token_hex(8)}',
            validateRequestBody=validate_request_body,
            validateRequestParameters=validate_request_parameters)
        return self.client.create_request_validator(**params)['id']

    @staticmethod
    def get_request_validator_params(request_validator):
        try:
            validator_name = request_validator.pop('name')
        except KeyError:  # not critical error
            validator_name = None
        request_validator_params = {}

        if ('validate_request_body', True) in request_validator.items():
            request_validator_params.update(
                {REQ_VALIDATOR_PARAM_VALIDATE_BODY: True,
                 REQ_VALIDATOR_PARAM_NAME: 'Validate body'})
        if ('validate_request_parameters', True) in request_validator.items():
            request_validator_params.update(
                {REQ_VALIDATOR_PARAM_VALIDATE_PARAMS: True,
                 REQ_VALIDATOR_PARAM_NAME: 'Validate query string '
                                           'parameters and headers'})

        # Need an additional check for validator name.
        # If the user wants to validate both the body and the parameters,
        # the number of 'request_validator_params' parameters will be equal
        # to three
        SETTED_PARAMETERS = 3

        if len(request_validator_params) == SETTED_PARAMETERS:
            request_validator_params.update(
                {REQ_VALIDATOR_PARAM_NAME: 'Validate body, query '
                                           'string parameters, '
                                           'and headers'})
        if validator_name:
            request_validator_params.update(
                {REQ_VALIDATOR_PARAM_NAME: validator_name})
        return request_validator_params

    def create_integration(self, api_id, resource_id, method, int_type,
                           integration_method=None, uri=None, credentials=None,
                           request_parameters=None, request_templates=None,
                           passthrough_behavior=None, cache_namespace=None,
                           cache_key_parameters=None):
        params = dict(restApiId=api_id, resourceId=resource_id, type=int_type,
                      httpMethod=method)
        if integration_method:
            params['integrationHttpMethod'] = integration_method
        if uri:
            params['uri'] = uri
        if credentials:
            params['credentials'] = credentials
        if request_parameters:
            params['requestParameters'] = request_parameters
        if request_templates:
            params['requestTemplates'] = request_templates
        if passthrough_behavior:
            params['passthroughBehavior'] = passthrough_behavior
        if cache_namespace:
            params['cacheNamespace'] = cache_namespace
        if cache_key_parameters:
            params['cacheKeyParameters'] = cache_key_parameters
        self.client.put_integration(**params)

    def create_lambda_integration(self, lambda_arn, api_id, resource_id,
                                  method, request_templates=None,
                                  passthrough_behavior=None,
                                  credentials=None,
                                  enable_proxy=False,
                                  cache_key_parameters=None,
                                  request_parameters=None):
        """ Create API Gateway integration with lambda by name.

        :type lambda_arn: str
        :type api_id: str
        :type resource_id: str
        :type method: str
        :type request_templates: dict
        :type passthrough_behavior: str
        :param passthrough_behavior: WHEN_NO_MATCH , WHEN_NO_TEMPLATES , NEVER
        :type lambda_region: str
        :type credentials: str
        :param credentials: role arn
        :type request_parameters: dict
        :param request_parameters: A key-value map specifying request parameters
         (path, query string, header)
        """
        uri = ('arn:aws:apigateway:{0}:lambda:path/2015-03-31/functions/{1}'
               '/invocations').format(self.region, lambda_arn)

        int_type = 'AWS_PROXY' if enable_proxy else 'AWS'

        params = dict(int_type=int_type, integration_method='POST',
                      method=method, passthrough_behavior='WHEN_NO_MATCH',
                      uri=uri, api_id=api_id, resource_id=resource_id)
        if credentials:
            params['credentials'] = credentials
        if passthrough_behavior:
            params['passthrough_behavior'] = passthrough_behavior
        if request_templates:
            params['request_templates'] = request_templates
        if cache_key_parameters:
            params['cache_key_parameters'] = cache_key_parameters
        if request_parameters:
            params['request_parameters'] = request_parameters
        self.create_integration(**params)

    def create_service_integration(self, acc_id, api_id, resource_id,
                                   method, integration_method, role,
                                   action, request_templates=None,
                                   passthrough_behavior=None,
                                   request_parameters=None):
        """
        Create API Gateway integration with AWS service.

        :type acc_id: str
        :param acc_id: Account id
        :type api_id: str
        :param api_id: Identifier of the RestApi
        :type resource_id: str
        :param resource_id: Request's resource ID
        :type method: str
        :param method: Request's HTTP method
        :type integration_method: str
        :param integration_method: Integration HTTP method
        :type role: str
        :param role: Execution role
        :type action: str
        :param action: using for URI, general template:
        {region}:{subdomain.service|service}:path|action/{service_api}
        :type request_templates: dict
        :param request_templates: A key-value map where content type is a key
        and the template is the value
        :type passthrough_behavior: str
        :param passthrough_behavior: WHEN_NO_MATCH , WHEN_NO_TEMPLATES , NEVER
        :type request_parameters: dict
        :param request_parameters: A key-value map specifying request parameters
         (path, query string, header)
        """
        uri = 'arn:aws:apigateway:{0}'.format(action)

        credentials = ApiGatewayConnection.get_service_integration_credentials(
            acc_id, role
        )

        params = dict(int_type='AWS', integration_method=integration_method,
                      method=method, passthrough_behavior='WHEN_NO_MATCH',
                      uri=uri, api_id=api_id, resource_id=resource_id,
                      credentials=credentials)

        if passthrough_behavior:
            params['passthrough_behavior'] = passthrough_behavior
        if request_templates:
            params['request_templates'] = request_templates
        if request_parameters:
            params['request_parameters'] = request_parameters
        self.create_integration(**params)

    @staticmethod
    def get_service_integration_credentials(acc_id, role):
        return 'arn:aws:iam::*:user/*' if role == 'caller_identity' \
            else 'arn:aws:iam::{0}:role/{1}'.format(acc_id, role)

    def create_mock_integration(self, api_id, resource_id, method,
                                request_templates=None,
                                passthrough_behavior=None):
        params = dict(int_type='MOCK', method=method, api_id=api_id,
                      resource_id=resource_id)
        if passthrough_behavior:
            params['passthrough_behavior'] = passthrough_behavior
        if request_templates:
            params['request_templates'] = request_templates
        self.create_integration(**params)

    def create_http_integration(self, api_id, resource_id,
                                method, integration_method, uri,
                                request_templates=None,
                                passthrough_behavior=None, enable_proxy=False):

        int_type = 'HTTP_PROXY' if enable_proxy else 'HTTP'

        params = dict(int_type=int_type, integration_method=integration_method,
                      method=method, passthrough_behavior='WHEN_NO_MATCH',
                      uri=uri, api_id=api_id, resource_id=resource_id)

        if passthrough_behavior:
            params['passthrough_behavior'] = passthrough_behavior
        if request_templates:
            params['request_templates'] = request_templates
        self.create_integration(**params)

    def create_integration_response(self, api_id, resource_id, method,
                                    status_code=None, selection_pattern=None,
                                    response_parameters=None,
                                    response_templates=None,
                                    enable_cors=False):
        """
        :type api_id: str
        :type resource_id: str
        :type method: str
        :param method: Specifies the method request's HTTP method type.
        :type status_code: str
        :type selection_pattern: str
        :type response_parameters: dict
        :type response_templates: dict
        :type enable_cors: bool
        """
        response_allow_origin = "response.header.Access-Control-Allow-Origin"
        method_allow_origin = "method.{0}".format(response_allow_origin)
        params = dict(restApiId=api_id, httpMethod=method, statusCode='200',
                      resourceId=resource_id,
                      responseTemplates={'application/json': ''})
        if enable_cors:
            params['responseParameters'] = {method_allow_origin: "\'*\'"}
        if selection_pattern:
            params['selectionPattern'] = selection_pattern
        if status_code:
            params['statusCode'] = status_code
        if response_parameters:
            if enable_cors:
                response_parameters[method_allow_origin] = "\'*\'"
            params['responseParameters'] = response_parameters
        if response_templates:
            params['responseTemplates'] = response_templates
        self.client.put_integration_response(**params)

    def create_method_response(self, api_id, resource_id, method,
                               status_code=None, response_parameters=None,
                               response_models=None, enable_cors=False):
        """
        :type api_id: str
        :type resource_id: str
        :type method: str
        :param method: Specifies the method request's HTTP method type.
        :type status_code: str
        :type response_parameters: dict
        :type response_models: dict
        :type enable_cors: bool
        """
        response_allow_origin = "response.header.Access-Control-Allow-Origin"
        method_allow_origin = "method.{0}".format(response_allow_origin)
        params = dict(restApiId=api_id, resourceId=resource_id,
                      httpMethod=method, statusCode='200',
                      responseModels={'application/json': 'Empty'})
        if enable_cors:
            params['responseParameters'] = {method_allow_origin: False}
        if status_code:
            params['statusCode'] = status_code
        if response_parameters:
            if enable_cors:
                response_parameters[method_allow_origin] = False
            params['responseParameters'] = response_parameters
        if response_models:
            params['responseModels'] = response_models
        self.client.put_method_response(**params)

    def enable_cors_for_resource(self, api_id, resource_id):
        """ When your API's resources receive requests from a domain other than
        the API's own domain, you must enable cross-origin resource sharing
        (CORS) for selected methods on the resource.

        :type api_id: str
        :type resource_id: str
        """
        self.create_method(api_id, resource_id, 'OPTIONS')
        self.create_integration(api_id=api_id, resource_id=resource_id,
                                method='OPTIONS', int_type='MOCK',
                                request_templates={
                                    'application/json': '{"statusCode": 200}'
                                })

        self.create_method_response(
            api_id, resource_id, 'OPTIONS', response_parameters={
                RESPONSE_PARAM_ALLOW_HEADERS: False,
                RESPONSE_PARAM_ALLOW_METHODS: False,
                RESPONSE_PARAM_ALLOW_ORIGIN: False
            })
        content_types = ("'Content-Type,X-Amz-Date,Authorization,X-Api-Key,"
                         "X-Amz-Security-Token'")
        self.create_integration_response(
            api_id, resource_id, 'OPTIONS', response_parameters={
                RESPONSE_PARAM_ALLOW_HEADERS: content_types,
                RESPONSE_PARAM_ALLOW_METHODS: "\'*\'",
                RESPONSE_PARAM_ALLOW_ORIGIN: "\'*\'"
            })

    def deploy_api(self, api_id, stage_name, stage_description='',
                   description='', cache_cluster_enabled=False,
                   cache_cluster_size=None, variables=None):
        """
        :type api_id: str
        :type stage_name: str
        :type stage_description: str
        :param stage_name:
        :param stage_description:
        :param description:
        :param cache_cluster_enabled:
        :param cache_cluster_size: '0.5'|'1.6'|'6.1'|'13.5'|'28.4'|'58.2'|'118'
        |'237'
        :type cache_cluster_size: str
        :type variables: dict
        """
        params = dict(restApiId=api_id, stageName=stage_name)
        if stage_description:
            params['stageDescription'] = stage_description
        if description:
            params['description'] = description
        if cache_cluster_enabled:
            params['cacheClusterEnabled'] = cache_cluster_enabled
        if cache_cluster_size:
            params['cacheClusterSize'] = cache_cluster_size
        if variables:
            params['variables'] = variables
        self.client.create_deployment(**params)

    def get_all_apis(self):
        """ Get all existing APIs."""
        existing_apis = []
        response = self.client.get_rest_apis()
        existing_apis.extend(response.get('items'))
        token = response.get('position')
        while token:
            response = self.client.get_rest_apis(position=token)
            existing_apis.extend(response.get('items'))
            token = response.get('position')
        return existing_apis

    def get_api(self, api_id):
        return self.client.get_rest_api(restApiId=api_id)

    def get_gateway_responses(self, api_id):
        return self.client.get_gateway_responses(restApiId=api_id).get('items',
                                                                       [])

    def add_header_to_gateway_response(self, api_id, response_type, name,
                                       value):
        path = '/responseParameters/gatewayresponse.header.' + name

        operation = {
            'op': 'add',
            'path': path,
            'value': value
        }
        _LOG.info(
            'Update Gateway response context: api_id:{0}; '
            'response_type:{1}; name:{2}; operation:{3};'.format(
                api_id, response_type, name, operation))

        return self.client.update_gateway_response(restApiId=api_id,
                                                   responseType=response_type,
                                                   patchOperations=[operation])

    def generate_sdk(self, api_id, stage_name='prod', sdk_type='javascript'):
        """ generates sdk of given type for specified stage of api gateway
        :param api_id: api gateway name
        :type api_id: str
        :param stage_name: stage name
        :type stage_name" str
        :param sdk_type: sdk type, possible values: javascript, android,
        objective c
        :type sdk_type: str
        """
        return self.client.get_sdk(restApiId=api_id, stageName=stage_name,
                                   sdkType=sdk_type)

    def update_configuration(self, rest_api_id, stage_name,
                             patch_operations):
        return self.client.update_stage(
            restApiId=rest_api_id,
            stageName=stage_name,
            patchOperations=patch_operations
        )

    def create_authorizer(self, api_id, name, type, provider_arns=None,
                          auth_type=None, authorizer_uri=None,
                          authorizer_credentials=None, identity_source=None,
                          validation_expression=None, ttl=None):
        params = dict(restApiId=api_id, name=name, type=type)
        if provider_arns:
            params['providerARNs'] = provider_arns
        if auth_type:
            params['authType'] = auth_type
        if authorizer_uri:
            params['authorizerUri'] = authorizer_uri
        if authorizer_credentials:
            params['authorizerCredentials'] = authorizer_credentials
        if identity_source:
            params['identitySource'] = identity_source
        if validation_expression:
            params['identityValidationExpression'] = validation_expression
        if ttl:
            params['authorizerResultTtlInSeconds'] = ttl

        return self.client.create_authorizer(**params)

    def get_authorizers(self, api_id):
        items = []
        params = dict(restApiId=api_id)
        response = self.client.get_authorizers(**params)
        if response.get('items'):
            items.extend(response.get('items'))
        position = response.get('position')
        while position:
            params['position'] = position
            response = self.client.get_authorizers(**params)
            if response.get('items'):
                items.extend(response.get('items'))
            position = response.get('position')
        return items

    def get_authorizer(self, rest_api_id, authorizer_id):
        return self.client.get_authorizer(restApiId=rest_api_id,
                                          authorizerId=authorizer_id)

    def update_compression_size(self, rest_api_id, compression_size=None):
        """Enables api compression and sets minimum compression size equal
        to given param. If the param wasn't given, will be disabled"""
        patchOperation = {
            'op': 'replace',
            'path': '/minimumCompressionSize',
        }
        if compression_size:
            patchOperation['value'] = str(compression_size)
            _LOG.debug(f'Setting compression size to "{compression_size}"')
        params = dict(restApiId=rest_api_id)
        params['patchOperations'] = [patchOperation, ]
        _LOG.debug(f'Updating rest api with params: "{params}"')
        return self.client.update_rest_api(**params)

    def create_model(self, rest_api_id, name, content_type, description=None,
                     schema=None):
        """Adds a new Model resource to an existing RestApi resource."""
        _LOG.debug(f'Creating new model "{name}"')
        params = {
            'contentType': content_type,
            'restApiId': rest_api_id,
            'name': name
        }
        if description:
            params['description'] = description
        if schema:
            params['schema'] = schema
        return self.client.create_model(**params)

    def delete_model(self, rest_api_id, name):
        _LOG.debug(f'Deleting model "{name}"')
        return self.client.delete_model(restApiId=rest_api_id, modelName=name)

    def get_model(self, rest_api_id, name, flatten=False):
        try:
            return self.client.get_model(restApiId=rest_api_id, modelName=name,
                                         flatten=flatten)
        except ClientError as e:
            if 'NotFoundException' in str(e):
                _LOG.warn(f'Cannot find model "{name}"')
                return None
            else:
                raise e


class ApiGatewayV2Connection:
    def __init__(self, region=None, aws_access_key_id=None,
                 aws_secret_access_key=None, aws_session_token=None):
        self._region = region
        self._aws_access_key_id = aws_access_key_id
        self._aws_secret_access_key = aws_secret_access_key
        self._aws_session_token = aws_session_token
        self._client: Optional[BaseClient] = None

    @property
    def client(self) -> BaseClient:
        if not self._client:
            self._client = client(
                'apigatewayv2', self._region,
                aws_access_key_id=self._aws_access_key_id,
                aws_secret_access_key=self._aws_secret_access_key,
                aws_session_token=self._aws_session_token
            )
            _LOG.debug('Opened new API Gateway connection.')
        return self._client

    def create_web_socket_api(self, name: str,
                              route_selection_expression: Optional[
                                  str] = 'request.body.action',
                              tags=None) -> str:
        params = dict(
            Name=name,
            ProtocolType='WEBSOCKET',
            RouteSelectionExpression=route_selection_expression
        )
        if tags:
            params['Tags'] = tags
        return self.client.create_api(**params)['ApiId']

    def create_stage(self, api_id: str, stage_name: str):
        return self.client.create_stage(
            ApiId=api_id,
            AutoDeploy=True,
            StageName=stage_name
        )

    # def create_deployment(self, api_id: str, stage_name: str):
    #     return self.client.create_deployment(
    #         ApiId=api_id,
    #         StageName=stage_name
    #     )

    def get_api_by_name(self, name: str) -> Optional[dict]:
        return next((api for api in self.client.get_apis()['Items']
                     if api['Name'] == name), None)

    def get_route_id_by_name(self, api_id: str, name: str) -> Optional[str]:
        return next(
            (route['RouteId'] for route in
             self.client.get_routes(ApiId=api_id)['Items']
             if route['RouteKey'] == name), None
        )

    def delete_api(self, api_id: str):
        return self.client.delete_api(
            ApiId=api_id
        )

    def create_lambda_integration(self, api_id: str, lambda_arn: str,
                                  enable_proxy: Optional[bool] = False) -> str:
        integration_uri = \
            f'arn:aws:apigateway:{self.client.meta.region_name}:lambda:path/' \
            f'2015-03-31/functions/{lambda_arn}/invocations'
        return self.client.create_integration(
            ApiId=api_id,
            IntegrationType='AWS_PROXY' if enable_proxy else 'AWS',
            IntegrationMethod='POST',
            IntegrationUri=integration_uri
        )['IntegrationId']

    def put_route_integration(self, api_id: str, route_name: str,
                              integration_id: str) -> str:
        route_id = self.get_route_id_by_name(api_id, route_name)
        if route_id:  # already exists
            self.client.update_route(
                ApiId=api_id,
                RouteId=route_id,
                Target=f'integrations/{integration_id}'
            )
        else:  # not route_id, does not exist
            response = self.client.create_route(
                ApiId=api_id,
                RouteKey=route_name,
                Target=f'integrations/{integration_id}'
            )
            route_id = response['RouteId']
        return route_id

    def get_routes(self, api_id: str):
        return self.client.get_routes(ApiId=api_id)

    def get_integration(self, api_id: str, integration_id: str):
        integration = self.client.get_integration(ApiId=api_id,
                                                  IntegrationId=integration_id)

        if integration.get('IntegrationMethod') == 'POST' and \
                integration.get('IntegrationUri'):
            return integration['IntegrationUri'].split('functions/')[-1].\
                replace('/invocations', '')



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from botocore.exceptions import ClientError

from syndicate.exceptions import ResourceNotFoundError, \
    ResourceProcessingError
from syndicate.commons.log_helper import get_logger, get_user_logger
from syndicate.core.helper import unpack_kwargs
from syndicate.core.resources.abstract_external_resource import AbstractExternalResource
from syndicate.core.resources.base_resource import BaseResource
from syndicate.core.resources.helper import (build_description_obj,
                                             validate_params)

AUTOSCALING_REQUIRED_PARAMS = ['resource_name', 'dimension',
                               'min_capacity', 'max_capacity',
                               'role_name']

DYNAMODB_TABLE_REQUIRED_PARAMS = ['hash_key_name', 'hash_key_type']

_LOG = get_logger(__name__)
USER_LOG = get_user_logger()


class DynamoDBResource(AbstractExternalResource, BaseResource):

    def __init__(self, dynamodb_conn, cw_alarm_conn,
                 app_as_conn, iam_conn) -> None:
        self.dynamodb_conn = dynamodb_conn
        self.cw_alarm_conn = cw_alarm_conn
        self.app_autoscaling_conn = app_as_conn
        self.iam_conn = iam_conn

    def create_tables(self, args, step=10):
        """ Only 10 tables can be created, updated or deleted simultaneously.

        :param args: list of tables configurations meta
        :type args: list
        :param step: how many tables to create simultaneously
        :type step: int
        :returns tables create results as list
        """
        return self.create_pool(job=self._create_dynamodb_table_from_meta,
                                parameters=args, workers=step)

    def update_tables(self, args, step=10):
        """ Only 10 tables can be created, updated or deleted simultaneously.

        :param args: list of tables configurations meta
        :type args: list
        :param step: how many tables to update simultaneously
        :type step: int
        :returns tables update results as list
        """
        return self.create_pool(job=self._update_dynamodb_table_from_meta,
                                parameters=args, workers=step)

    def remove_dynamodb_tables(self, args, step=10):
        return self.create_pool(self._remove_tables_from_meta, args,
                                workers=step)

    @unpack_kwargs
    def _update_dynamodb_table_from_meta(self, name, meta, context):
        """ Update Dynamo DB table from meta description, specifically:
        capacity (billing) mode, table or gsi capacity units,
        gsi to create or delete, ttl.

        :param name: DynamoDB table name
        :type name: str
        :param meta: table configuration information
        :type meta: dict
        :returns table update result as dict
        """
        table = self.dynamodb_conn.get_table_by_name(name)
        if not table:
            raise ResourceNotFoundError(f'{name} table does not exist.')

        billing_mode = meta.get('billing_mode', 'PROVISIONED')

        read_capacity = table.provisioned_throughput.get('ReadCapacityUnits')
        write_capacity = table.provisioned_throughput.get('WriteCapacityUnits')

        if read_capacity and write_capacity:
            existing_billing_mode = 'PROVISIONED'
            provisioned_throughput = dict(
                ReadCapacityUnits=read_capacity,
                WriteCapacityUnits=write_capacity
            )
            on_demand_throughput = dict()
        else:
            existing_billing_mode = 'PAY_PER_REQUEST'
            on_demand_throughput = dict(
                MaxReadRequestUnits=table.on_demand_throughput.get(
                    'MaxReadRequestUnits'),
                MaxWriteRequestUnits=table.on_demand_throughput.get(
                    'MaxWriteRequestUnits'
                )
            )
            provisioned_throughput = dict()

        if billing_mode != existing_billing_mode:
            USER_LOG.info(
                f"Updating the table '{name}'. It may take up to 20 minutes, "
                f"please wait.")

        response = self.dynamodb_conn.update_table_capacity(
            table_name=name,
            billing_mode=billing_mode,
            read_capacity=meta.get('read_capacity'),
            write_capacity=meta.get('write_capacity'),
            existing_billing_mode=existing_billing_mode,
            existing_provisioned_throughput=provisioned_throughput,
            existing_on_demand_throughput=on_demand_throughput,
            existing_global_indexes=table.global_secondary_indexes or []
        )
        if response:
            table = response

        self.dynamodb_conn.update_table_ttl(
            table_name=name,
            ttl_attribute_name=meta.get('ttl_attribute_name')
        )

        table_read_capacity = \
            read_capacity or on_demand_throughput.get('MaxReadRequestUnits')
        table_write_capacity = \
            write_capacity or on_demand_throughput.get('MaxWriteRequestUnits')

        global_indexes_meta = meta.get('global_indexes', [])
        self.dynamodb_conn.update_global_indexes(
            table_name=name,
            global_indexes_meta=global_indexes_meta,
            existing_global_indexes=table.global_secondary_indexes or [],
            table_read_capacity=table_read_capacity,
            table_write_capacity=table_write_capacity,
            existing_capacity_mode=existing_billing_mode
        )

        return self.describe_table(name, meta)

    def describe_table(self, name, meta, response=None):
        if not response:
            response = self.dynamodb_conn.describe_table(table_name=name)
        if not response:
            return {}
        arn = response['TableArn']
        del response['TableArn']
        return {
            arn: build_description_obj(response, name, meta)
        }

    def describe_stream(self, name, meta):
        response = self.dynamodb_conn.describe_table(meta['table'])
        if not response:
            return {}
        res_obj = {
            'StreamSpecification': response['StreamSpecification'],
            'LatestStreamLabel': response['LatestStreamLabel']
        }
        arn = response['LatestStreamArn']
        return {
            arn: build_description_obj(res_obj, name, meta)
        }

    def describe_meta(self, name):
        meta = {
            'resource_type': 'dynamodb_table',
            'external': True
        }
        response = self.dynamodb_conn.describe_table(table_name=name)

        if not response:
            return {}

        key_schema = {k['KeyType']: k['AttributeName'] for k in response['KeySchema']}
        attribute_definitions = {k['AttributeName']: k['AttributeType'] for k in response['AttributeDefinitions']}

        hash_key_name = key_schema.get('HASH')
        hash_key_type = attribute_definitions[hash_key_name] if hash_key_name else None

        sort_key_name = key_schema.get('RANGE')
        sort_key_type = attribute_definitions[sort_key_name] if sort_key_name else None

        meta.update({
            'hash_key_name': hash_key_name,
            'hash_key_type': hash_key_type,
            'sort_key_name': sort_key_name,
            'sort_key_type': sort_key_type,
        })

        gsi = response.get('GlobalSecondaryIndexes')
        if gsi:
            global_indexes = []
            for index in gsi:
                key_schema = {k['KeyType']: k['AttributeName'] for k in index['KeySchema']}

                index_key_name = key_schema.get('HASH')
                index_key_type = attribute_definitions.get(index_key_name) if index_key_name else None

                sort_key_name = key_schema.get('RANGE')
                sort_key_type = attribute_definitions.get(sort_key_name) if sort_key_name else None

                index_data = {
                    'name': index.get('IndexName'),
                    'index_key_name': index_key_name,
                    'index_key_type': index_key_type,
                    'index_sort_key_name': sort_key_name,
                    'index_sort_key_type': sort_key_type,
                }
                global_indexes.append(index_data)
            global_indexes.sort(key=lambda k:k['name'])
            meta['global_indexes'] = global_indexes

        return {name: meta}

    def define_resource_shape(self):
        return {
            'resource_type': None,
            'hash_key_name': None,
            'hash_key_type': None,
            'sort_key_name': None,
            'sort_key_type': None,
            'global_indexes': [
                {
                    'name': None,
                    'index_key_name': None,
                    'index_key_type': None,
                    'sort_key_name': None,
                    'sort_key_type': None,
                    'index_sort_key_name': None,
                    'index_sort_key_type': None
                }
            ]
        }

    @unpack_kwargs
    def _create_dynamodb_table_from_meta(self, name, meta):
        """ Create Dynamo DB table from meta description after parameter
        validation.
    
        :type name: str
        :type meta: dict
        """
        validate_params(name, meta, DYNAMODB_TABLE_REQUIRED_PARAMS)

        res = self.dynamodb_conn.describe_table(name)
        autoscaling_config = meta.get('autoscaling')
        if res:
            _LOG.warn('%s table exists.', name)
            if autoscaling_config:
                res['Autoscaling'] = self._describe_autoscaling(
                    autoscaling_config,
                    name)
            response = self.describe_table(name, meta, res)
            return response

        self.dynamodb_conn.create_table(
            name, meta['hash_key_name'], meta['hash_key_type'],
            meta.get('billing_mode', 'PROVISIONED'),
            meta.get('sort_key_name'), meta.get('sort_key_type'),
            meta.get('read_capacity'), meta.get('write_capacity'),
            global_indexes=meta.get('global_indexes'),
            local_indexes=meta.get('local_indexes'),
            tags=meta.get('tags'),
            wait=True)
        response = self.dynamodb_conn.describe_table(name)
        if not response:
            raise ResourceProcessingError(
                f"Table with name '{name}' has not been created!"
            )
        # enabling stream if present
        stream_view_type = meta.get('stream_view_type')
        if stream_view_type:
            stream = self.dynamodb_conn.get_table_stream_arn(name)
            if stream:
                _LOG.warn('Stream %s exists.', name)
            else:
                try:
                    self.dynamodb_conn.enable_table_stream(name,
                                                           stream_view_type)
                except ClientError as e:
                    # handle specific case for fantom stream enabling
                    if 'ResourceInUseException' in str(e):
                        _LOG.warn('Stream enabling currently in progress,'
                                  ' table: %s', name)
                    else:
                        raise e
        if autoscaling_config:
            _LOG.debug('Found autoscaling configuration for resource %s', name)
            sc_res = self._enable_autoscaling(autoscaling_config, name)
            response['Autoscaling'] = sc_res
        _LOG.info('Created table %s.', name)
        return self.describe_table(name, meta, response)

    def _describe_autoscaling(self, autoscaling_config, name):
        targets = []
        policies = []
        for item in autoscaling_config:
            dimension = item['dimension']
            resource_name = item['resource_name']
            resource_id = self.build_res_id(dimension, resource_name, name)
            sc_targets = self.app_autoscaling_conn.describe_scalable_targets(
                service_namespace='dynamodb',
                resources_ids=[resource_id],
                scalable_dimension=dimension)
            targets.extend(sc_targets)
            autoscaling_policy = item.get('config')
            if autoscaling_policy:
                policy_name = autoscaling_policy['policy_name']
                sc_policies = self.app_autoscaling_conn.describe_scaling_policies(
                    service_namespace='dynamodb', policy_names=[policy_name],
                    resource_id=resource_id, scalable_dimension=dimension)
                policies.extend(sc_policies)
        return {
            'targets': targets,
            'policies': policies
        }

    def _enable_autoscaling(self, autoscaling_config, name):
        targets = []
        policies = []
        for item in autoscaling_config:
            validate_params(name, item, AUTOSCALING_REQUIRED_PARAMS)
            role_name = item['role_name']
            role_arn = self.iam_conn.check_if_role_exists(role_name)
            if role_arn:
                dimension = item['dimension']
                resource_id, sc_targets = self.register_autoscaling_target(
                    dimension,
                    item,
                    role_arn,
                    name)
                targets.extend(sc_targets)
                _LOG.debug('Autoscaling %s is set up for %s', dimension,
                           resource_id)
                autoscaling_policy = item.get('config')
                if autoscaling_policy:
                    policy_name = autoscaling_policy['policy_name']
                    _LOG.debug('Going to set up autoscaling with '
                               'policy %s', policy_name)
                    sc_policies = self.put_autoscaling_policy(
                        autoscaling_policy,
                        dimension,
                        policy_name,
                        resource_id)
                    policies.append(sc_policies)
                    _LOG.debug('Policy %s is set up', policy_name)
            else:
                _LOG.warn('Role %s is not found, skip autoscaling config',
                          role_name)
        return {
            'targets': targets,
            'policies': policies
        }

    def put_autoscaling_policy(self, autoscaling_policy, dimension,
                               policy_name,
                               resource_id):
        target_utilization = autoscaling_policy['target_utilization']
        scale_in_cooldown = autoscaling_policy.get('scale_in_cooldown')
        scale_out_cooldown = autoscaling_policy.get('scale_out_cooldown')
        metric_type = 'DynamoDBWriteCapacityUtilization' \
            if 'Write' in dimension \
            else 'DynamoDBReadCapacityUtilization'
        response = self.app_autoscaling_conn.put_target_scaling_policy(
            policy_name=policy_name, service_namespace='dynamodb',
            resource_id=resource_id, scalable_dimension=dimension,
            target_value=target_utilization,
            predefined_metric_type=metric_type,
            scale_in_cooldown=scale_in_cooldown,
            scale_out_cooldown=scale_out_cooldown)
        return response

    def register_autoscaling_target(self, dimension, item, role_arn,
                                    table_name):
        resource_name = item['resource_name']
        resource_id = self.build_res_id(dimension, resource_name, table_name)
        self.app_autoscaling_conn.register_target(service_namespace='dynamodb',
                                                  resource_id=resource_id,
                                                  scalable_dimension=dimension,
                                                  min_capacity=str(
                                                      item['min_capacity']),
                                                  max_capacity=str(
                                                      item['max_capacity']),
                                                  role_arn=role_arn)
        targets = self.app_autoscaling_conn.describe_scalable_targets(
            service_namespace='dynamodb',
            resources_ids=[resource_id],
            scalable_dimension=dimension)
        return resource_id, targets

    @staticmethod
    def build_res_id(dimension, resource_name, table_name):
        resource_id = f'table/{table_name}' if 'table' in dimension \
            else f'table/{table_name}/index/{resource_name}'
        return resource_id

    @unpack_kwargs
    def _remove_tables_from_meta(self, arn, config):
        db_name = config['resource_name']
        removed_tables, errors = self.dynamodb_conn.remove_tables_by_names(
            [db_name],
            log_not_found_error=False)
        if errors:
            raise ResourceProcessingError('; '.join(errors))
        _LOG.info(f'Dynamo DB tables {str(removed_tables)} were removed')

        alarm_args = []
        autoscaling = config['description'].get('Autoscaling')
        if autoscaling:
            policies = autoscaling['policies']
            for policy in policies:
                if policy:
                    alarms = policy.get('Alarms', [])
                    alarm_args.extend(map(lambda x: x['AlarmName'], alarms))
        try:
            if alarm_args:
                self.cw_alarm_conn.remove_alarms(alarm_args)
        except Exception as e:
            raise e

        return {arn: config}



"""
    Copyright 2018 EPAM Systems, Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import os
from functools import partial

import click

from syndicate.commons.log_helper import get_user_logger
from syndicate.core.conf.generator import generate_configuration_files
from syndicate.core.constants import SYNDICATE_WIKI_PAGE, FAILED_RETURN_CODE, \
    JAVA_LAMBDAS_WIKI_PAGE, SYNDICATE_PROJECT_EXAMPLES_PAGE, OK_RETURN_CODE, \
    ABORTED_RETURN_CODE
from syndicate.core.decorators import return_code_manager
from syndicate.core.generators.lambda_function import (
    generate_lambda_function, generate_lambda_layer)
from syndicate.core.generators.project import (generate_project_structure,
                                               PROJECT_PROCESSORS)
from syndicate.core.generators.swagger_ui import generate_swagger_ui
from syndicate.core.groups import RUNTIME_JAVA
from syndicate.core.groups.appsync import appsync
from syndicate.core.groups.meta import meta
from syndicate.core.helper import timeit, validate_bucket_name, \
    resolve_project_path, check_lambdas_names, DictParamType, check_suffix, \
    check_prefix, check_file_extension, check_lambda_layer_name, check_tags, \
    check_lambda_existence, verbose_option, AliasedCommandsGroup, \
    MultiWordOption, resolve_deploy_target_bucket_param

GENERATE_GROUP_NAME = 'generate'
GENERATE_PROJECT_COMMAND_NAME = 'project'
GENERATE_CONFIG_COMMAND_NAME = 'config'


USER_LOG = get_user_logger()


@click.group(name=GENERATE_GROUP_NAME, cls=AliasedCommandsGroup)
def generate():
    """Generates project, lambda or configs"""


@generate.command(name=GENERATE_PROJECT_COMMAND_NAME)
@return_code_manager
@click.option('--name', nargs=1, required=True, help='The project name')
@click.option('--path', nargs=1,
              help='Path to folder where the project will be created. '
                   'Default value: current working directory')
@verbose_option
@timeit()
def project(name, path):
    """
    Generates project with all the necessary components and in a right
    folders/files hierarchy to start developing in a min.
    """
    USER_LOG.info(f'Project name: {name}')

    proj_path = os.getcwd() if not path else path
    if not os.access(proj_path, os.X_OK | os.W_OK):
        USER_LOG.error(
            f"Incorrect permissions for the provided path '{proj_path}'")
        return FAILED_RETURN_CODE
    USER_LOG.info(f'Project path: {proj_path}')
    generate_project_structure(project_name=name,
                               project_path=proj_path)
    return OK_RETURN_CODE


@generate.command(name='lambda')
@return_code_manager
@click.option('--name', multiple=True, type=str,
              required=True, callback=check_lambdas_names,
              help='(multiple) The lambda function name')
@click.option('--runtime', required=True,
              help='Lambda\'s runtime. If multiple lambda names are specified,'
                   ' the runtime will be applied to all lambdas',
              type=click.Choice(PROJECT_PROCESSORS))
@click.option('--project-path', '-path', cls=MultiWordOption,
              help="Path to the project root directory. Default value: "
                   "the one from the current config if it exists. "
                   "Otherwise - the current working directory",
              nargs=1, callback=resolve_project_path)
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='The resource tags')
@verbose_option
@timeit()
def lambda_function(name, runtime, project_path, tags):
    """
    Generates required environment for lambda function
    """
    tags = tags or {}
    if not os.access(project_path, os.F_OK):
        USER_LOG.error(f"The provided path {project_path} doesn't exist")
        return FAILED_RETURN_CODE
    elif not os.access(project_path, os.W_OK) or not os.access(project_path,
                                                               os.X_OK):
        USER_LOG.error(f"Incorrect permissions for the provided path "
                       f"'{project_path}'")
        return FAILED_RETURN_CODE
    USER_LOG.info(f'Lambda names: {name}')
    USER_LOG.info(f'Runtime: {runtime}')
    USER_LOG.info(f'Project path: {project_path}')
    generate_lambda_function(project_path=project_path,
                             runtime=runtime,
                             lambda_names=name,
                             tags=tags)
    return OK_RETURN_CODE


@generate.command(name='lambda-layer')
@return_code_manager
@click.option('--name', type=str, required=True,
              callback=check_lambda_layer_name,
              help='The lambda layer name')
@click.option('--runtime', required=True,
              type=click.Choice(PROJECT_PROCESSORS),
              help='Lambda layer\'s runtime.')
@click.option('--link-with-lambda', cls=MultiWordOption,
              required=False,
              type=str, multiple=True, callback=check_lambda_existence,
              help='(multiple) Lambda function name to link the layer with.')
@click.option('--project-path', '-path', cls=MultiWordOption, nargs=1,
              help="Path to the project root directory. Default value: "
                   "the one from the current config if it exists. "
                   "Otherwise - the current working directory",
              callback=resolve_project_path)
@verbose_option
@timeit()
def lambda_layer(name, runtime, link_with_lambda, project_path):
    """
    Generates required environment for lambda function's layer
    """
    if runtime == RUNTIME_JAVA:
        USER_LOG.warning(
            'Generation of lambda layer for Java runtime is currently '
            'unsupported. \nA layer for lambda with Java runtime can '
            'be added to the project by using the annotation '
            '\'@LambdaLayer\'. \nMore details can be found on the '
            'aws-syndicate wiki page or in the project examples:\n'
            f'{SYNDICATE_WIKI_PAGE + JAVA_LAMBDAS_WIKI_PAGE}'
            f'\n{SYNDICATE_PROJECT_EXAMPLES_PAGE + runtime}'
        )
        return ABORTED_RETURN_CODE
    if not os.access(project_path, os.F_OK):
        USER_LOG.error(f"The provided path {project_path} doesn't exist")
        return FAILED_RETURN_CODE
    elif not os.access(project_path, os.W_OK) or not os.access(project_path,
                                                               os.X_OK):
        USER_LOG.error(f"Incorrect permissions for the provided path "
                       f"'{project_path}'")
        return FAILED_RETURN_CODE
    USER_LOG.info(f'Project path: {project_path}')
    generate_lambda_layer(name=name,
                          runtime=runtime,
                          lambda_names=link_with_lambda,
                          project_path=project_path)
    return OK_RETURN_CODE


@generate.command(name=GENERATE_CONFIG_COMMAND_NAME)
@return_code_manager
@click.option('--name',
              required=True,
              help='Name of the configuration to create. '
                   'Generated config will be created in folder '
                   '.syndicate-config-{name}. May contain name '
                   'of the environment')
@click.option('--region',
              help='The region that is used to deploy the application',
              required=True)
@click.option('--bundle-bucket-name', cls=MultiWordOption,
              is_eager=True, hidden=True,
              ) # this parameter is here for backward compatibility with versions lower than 1.18.0
@click.option('--deploy-target-bucket', cls=MultiWordOption,
              help="Name of the bucket that is used for uploading artifacts. "
                   "To create it, execute the command "
                   "'syndicate create-deploy-target-bucket'  [required]",
              callback=resolve_deploy_target_bucket_param)
@click.option('--access-key', cls=MultiWordOption,
              help='AWS access key id that is used to deploy the application. '
                   'Retrieved from session by default')
@click.option('--secret-key', cls=MultiWordOption,
              help='AWS secret key that is used to deploy the application. '
                   'Retrieved from session by default')
@click.option('--session-token', cls=MultiWordOption,
              help='AWS session token that is used to deploy the application. '
                   'Retrieved from session by default')
@click.option('--config-path', cls=MultiWordOption,
              help='Path to store generated configuration file')
@click.option('--project-path', '-path', cls=MultiWordOption,
              help='Path to the project root directory. '
                   'Default value: current working directory')
@click.option('--prefix',
              help='Prefix that is added to project names while deployment '
                   'by pattern: {prefix}resource_name{suffix}. '
                   'Must be less than or equal to 5. If --extended_prefix '
                   'specified prefix length may be up to 14 symbols',
              callback=check_prefix)
@click.option('--suffix',
              help='Suffix that is added to project names while deployment '
                   'by pattern: {prefix}resource_name{suffix}. '
                   'Must be less than or equal to 5',
              callback=check_suffix)
@click.option('--extended-prefix', cls=MultiWordOption,
              type=bool, default=False, is_eager=True,
              help='Extends the length of the prefix up to 14 symbols. '
                   'If specified, a prefix and a suffix will be added to all '
                   'project resources.')
@click.option('--use-temp-creds', cls=MultiWordOption, type=bool,
              default=False,
              help='Indicates Syndicate to generate and use temporary AWS '
                   'credentials')
@click.option('--access-role', cls=MultiWordOption, type=str,
              help='Indicates Syndicate to use this role\'s temporary AWS '
                   'credentials. Cannot be used if \'--use_temp_creds\' is '
                   'equal to true')
@click.option('--serial-number', cls=MultiWordOption, type=str,
              help='The identification number of the MFA device that is '
                   'associated with the IAM user which will be used for '
                   'deployment. If specified MFA token will be asked before '
                   'making actions')
@click.option('--tags', type=DictParamType(), callback=check_tags,
              help='Tags to add to the config. They will be added to all the '
                   'resources during deployment')
@click.option('--iam-permissions-boundary', cls=MultiWordOption,
              type=str,
              help='Common permissions boundary arn to add to all the roles')
@click.option('--lock-lifetime-minutes', cls=MultiWordOption,
              type=click.IntRange(min=0, max=300),
              help='Project lock lifetime in minutes (see section "locks" '
                   'in ".syndicate"). The default value is 20 minutes.')
@verbose_option
@timeit()
def config(name, config_path, project_path, region, access_key, secret_key,
           session_token, deploy_target_bucket, prefix, suffix,
           extended_prefix, use_temp_creds, access_role, serial_number,
           tags, iam_permissions_boundary, lock_lifetime_minutes,
           bundle_bucket_name):
    """
    Creates Syndicate configuration files
    """
    generate_configuration_files(name=name,
                                 config_path=config_path,
                                 project_path=project_path,
                                 region=region,
                                 access_key=access_key,
                                 secret_key=secret_key,
                                 session_token=session_token,
                                 deploy_target_bucket=deploy_target_bucket,
                                 prefix=prefix,
                                 suffix=suffix,
                                 extended_prefix=extended_prefix,
                                 use_temp_creds=use_temp_creds,
                                 access_role=access_role,
                                 serial_number=serial_number,
                                 tags=tags,
                                 iam_permissions_boundary=iam_permissions_boundary,
                                 lock_lifetime_minutes=lock_lifetime_minutes)
    return OK_RETURN_CODE


@generate.command(name='swagger-ui')
@return_code_manager
@click.option('--name', required=True, type=str,
              help="Swagger UI name")
@click.option('--path-to-spec', cls=MultiWordOption,
              required=True, type=str,
              callback=partial(check_file_extension, extensions=['.json']),
              help="Path to OpenAPI specification file. Path that is relative "
                   "to the project path can be specified.")
@click.option('--target-bucket', cls=MultiWordOption,
              required=True, type=str, callback=validate_bucket_name,
              help="S3 bucket name for Swagger UI deployment")
@click.option('--project-path', '-path', cls=MultiWordOption,
              help="Path to the project root directory. Default value: "
                   "the one from the current config if it exists. "
                   "Otherwise - the current working directory",
              nargs=1, callback=resolve_project_path)
@verbose_option
@timeit()
def swagger_ui(name, path_to_spec, target_bucket, project_path):
    """
    Generates required environment for Swagger UI
    """
    if not os.access(project_path, os.F_OK):
        USER_LOG.error(f"The provided path {project_path} doesn't exist")
        return FAILED_RETURN_CODE
    elif not os.access(project_path, os.W_OK) or not os.access(project_path,
                                                               os.X_OK):
        USER_LOG.error(f"Incorrect permissions for the provided path "
                       f"'{project_path}'")
        return FAILED_RETURN_CODE
    generate_swagger_ui(name=name,
                        spec_path=path_to_spec,
                        target_bucket=target_bucket,
                        project_path=project_path)
    return OK_RETURN_CODE


generate.add_command(meta)
generate.add_command(appsync)
